import cv2
import numpy as np
from copy import copy
from scipy.sparse import lil_matrix, dok_matrix
from scipy import optimize
from scipy import signal
from numba import jit
from collections import defaultdict, Counter
import toml
import itertools
from tqdm import trange
from pprint import pprint
import time

from aniposelib.boards import merge_rows, extract_points, \
    extract_rtvecs, get_video_params
from aniposelib.utils import get_initial_extrinsics, make_M, get_rtvec, \
    get_connections

@jit(nopython=True, parallel=True)
def triangulate_simple(points, camera_mats):
    num_cams = len(camera_mats)
    A = np.zeros((num_cams * 2, 4))
    for i in range(num_cams):
        x, y = points[i]
        mat = camera_mats[i]
        A[(i * 2):(i * 2 + 1)] = x * mat[2] - mat[0]
        A[(i * 2 + 1):(i * 2 + 2)] = y * mat[2] - mat[1]
    u, s, vh = np.linalg.svd(A, full_matrices=True)
    p3d = vh[-1]
    p3d = p3d[:3] / p3d[3]
    return p3d


def get_error_dict(errors_full, min_points=10):
    n_cams = errors_full.shape[0]
    errors_norm = np.linalg.norm(errors_full, axis=2)

    good = ~np.isnan(errors_full[:, :, 0])

    error_dict = dict()

    for i in range(n_cams):
        for j in range(i+1, n_cams):
            subset = good[i] & good[j]
            err_subset = errors_norm[:, subset][[i, j]]
            err_subset_mean = np.mean(err_subset, axis=0)
            if np.sum(subset) > min_points:
                percents = np.percentile(err_subset_mean, [15, 75])
                # percents = np.percentile(err_subset, [25, 75])
                error_dict[(i, j)] = (err_subset.shape[1], percents)
    return error_dict

def check_errors(cgroup, imgp):
    p3ds = cgroup.triangulate(imgp)
    errors_full = cgroup.reprojection_error(p3ds, imgp, mean=False)
    return get_error_dict(errors_full)

def subset_extra(extra, ixs):
    if extra is None:
        return None

    new_extra = {
        'objp': extra['objp'][ixs],
        'ids': extra['ids'][ixs],
        'rvecs': extra['rvecs'][:, ixs],
        'tvecs': extra['tvecs'][:, ixs]
    }
    return new_extra

def resample_points_extra(imgp, extra, n_samp=25):
    n_cams, n_points, _ = imgp.shape
    ids = remap_ids(extra['ids'])
    n_ids = np.max(ids)+1
    good = ~np.isnan(imgp[:, :, 0])
    ixs = np.arange(n_points)

    cam_counts = np.zeros((n_ids, n_cams), dtype='int64')
    for idnum in range(n_ids):
        cam_counts[idnum] = np.sum(good[:, ids == idnum], axis=1)
    cam_counts_random = cam_counts + np.random.random(size=cam_counts.shape)
    best_boards = np.argsort(-cam_counts_random, axis=0)

    cam_totals = np.zeros(n_cams, dtype='int64')

    include = set()
    for cam_num in range(n_cams):
        for board_id in best_boards[:, cam_num]:
            include.update(ixs[ids == board_id])
            cam_totals += cam_counts[board_id]
            if cam_totals[cam_num] >= n_samp or \
               cam_counts_random[board_id, cam_num] < 1:
                break

    final_ixs = sorted(include)
    newp = imgp[:, final_ixs]
    extra = subset_extra(extra, final_ixs)
    return newp, extra

def resample_points(imgp, extra=None, n_samp=25):
    # if extra is not None:
    #     return resample_points_extra(imgp, extra, n_samp)

    n_cams = imgp.shape[0]
    good = ~np.isnan(imgp[:, :, 0])
    ixs = np.arange(imgp.shape[1])

    num_cams = np.sum(~np.isnan(imgp[:, :, 0]), axis=0)

    include = set()

    for i in range(n_cams):
        for j in range(i+1, n_cams):
            subset = good[i] & good[j]
            n_good = np.sum(subset)
            if n_good > 0:
                ## pick points, prioritizing points seen by more cameras
                arr = np.copy(num_cams[subset]).astype('float64')
                arr += np.random.random(size=arr.shape)
                picked_ix = np.argsort(-arr)[:n_samp]
                picked = ixs[subset][picked_ix]
                include.update(picked)

    final_ixs = sorted(include)
    newp = imgp[:, final_ixs]
    extra = subset_extra(extra, final_ixs)
    return newp, extra

def medfilt_data(values, size=15):
    padsize = size+5
    vpad = np.pad(values, (padsize, padsize), mode='reflect')
    vpadf = signal.medfilt(vpad, kernel_size=size)
    return vpadf[padsize:-padsize]

def nan_helper(y):
    return np.isnan(y), lambda z: z.nonzero()[0]

def interpolate_data(vals):
    nans, ix = nan_helper(vals)
    out = np.copy(vals)
    try:
        out[nans] = np.interp(ix(nans), ix(~nans), vals[~nans])
    except ValueError:
        out[:] = 0
    return out

def remap_ids(ids):
    unique_ids = np.unique(ids)
    ids_out = np.copy(ids)
    for i, num in enumerate(unique_ids):
        ids_out[ids == num] = i
    return ids_out

def transform_points(points, rvecs, tvecs):
    """Rotate points by given rotation vectors and translate.
    Rodrigues' rotation formula is used.
    """
    theta = np.linalg.norm(rvecs, axis=1)[:, np.newaxis]
    with np.errstate(invalid='ignore'):
        v = rvecs / theta
        v = np.nan_to_num(v)
    dot = np.sum(points * v, axis=1)[:, np.newaxis]
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)

    rotated = cos_theta * points + \
        sin_theta * np.cross(v, points) + \
        dot * (1 - cos_theta) * v

    return rotated + tvecs


class Camera:
    def __init__(self,
                 matrix=np.eye(3),
                 dist=np.zeros(5),
                 size=None,
                 rvec=np.zeros(3),
                 tvec=np.zeros(3),
                 name=None,
                 extra_dist=False):

        self.set_camera_matrix(matrix)
        self.set_distortions(dist)
        self.set_size(size)
        self.set_rotation(rvec)
        self.set_translation(tvec)
        self.set_name(name)
        self.extra_dist = extra_dist

    def get_dict(self):
        return {
            'name': self.get_name(),
            'size': list(self.get_size()),
            'matrix': self.get_camera_matrix().tolist(),
            'distortions': self.get_distortions().tolist(),
            'rotation': self.get_rotation().tolist(),
            'translation': self.get_translation().tolist(),
        }

    def load_dict(self, d):
        self.set_camera_matrix(d['matrix'])
        self.set_rotation(d['rotation'])
        self.set_translation(d['translation'])
        self.set_distortions(d['distortions'])
        self.set_name(d['name'])
        self.set_size(d['size'])

    def from_dict(d):
        cam = Camera()
        cam.load_dict(d)
        return cam

    def get_camera_matrix(self):
        return self.matrix

    def get_distortions(self):
        return self.dist

    def set_camera_matrix(self, matrix):
        self.matrix = np.array(matrix, dtype='float64')

    def set_focal_length(self, fx, fy=None):
        if fy is None:
            fy = fx
        self.matrix[0, 0] = fx
        self.matrix[1, 1] = fy

    def get_focal_length(self, both=False):
        fx = self.matrix[0, 0]
        fy = self.matrix[1, 1]
        if both:
            return (fx, fy)
        else:
            return (fx + fy) / 2.0

    def set_distortions(self, dist):
        self.dist = np.array(dist, dtype='float64').ravel()

    def set_rotation(self, rvec):
        self.rvec = np.array(rvec, dtype='float64').ravel()

    def get_rotation(self):
        return self.rvec

    def set_translation(self, tvec):
        self.tvec = np.array(tvec, dtype='float64').ravel()

    def get_translation(self):
        return self.tvec

    def get_extrinsics_mat(self):
        return make_M(self.rvec, self.tvec)

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = str(name)

    def set_size(self, size):
        """set size as (width, height)"""
        self.size = size

    def get_size(self):
        """get size as (width, height)"""
        return self.size

    def resize_camera(self, scale):
        """resize the camera by scale factor, updating intrinsics to match"""
        size = self.get_size()
        new_size = size[0] * scale, size[1] * scale
        matrix = self.get_camera_matrix()
        new_matrix = matrix * scale
        new_matrix[2, 2] = 1
        self.set_size(new_size)
        self.set_camera_matrix(new_matrix)

    def get_params(self):
        params = np.zeros(8 + self.extra_dist, dtype='float64')
        params[0:3] = self.get_rotation()
        params[3:6] = self.get_translation()
        params[6] = self.get_focal_length()
        dist = self.get_distortions()
        params[7] = dist[0]
        if self.extra_dist:
            params[8] = dist[1]
        return params

    def set_params(self, params):
        self.set_rotation(params[0:3])
        self.set_translation(params[3:6])
        self.set_focal_length(params[6])

        dist = np.zeros(5, dtype='float64')
        dist[0] = params[7]
        if self.extra_dist:
            dist[1] = params[8]
        self.set_distortions(dist)

    def distort_points(self, points):
        shape = points.shape
        points = points.reshape(-1, 1, 2)
        new_points = np.dstack([points, np.ones((points.shape[0], 1, 1))])
        out, _ = cv2.projectPoints(new_points, np.zeros(3), np.zeros(3),
                                   self.matrix.astype('float64'),
                                   self.dist.astype('float64'))
        return out.reshape(shape)

    def undistort_points(self, points):
        shape = points.shape
        points = points.reshape(-1, 1, 2)
        out = cv2.undistortPoints(points,
                                  self.matrix.astype('float64'),
                                  self.dist.astype('float64'))
        return out.reshape(shape)

    def project(self, points):
        points = points.reshape(-1, 1, 3)
        out, _ = cv2.projectPoints(points, self.rvec, self.tvec,
                                   self.matrix.astype('float64'),
                                   self.dist.astype('float64'))
        return out

    def reprojection_error(self, p3d, p2d):
        proj = self.project(p3d).reshape(p2d.shape)
        return p2d - proj

    def copy(self):
        return \
            Camera(matrix=self.get_camera_matrix().copy(),
                   dist=self.get_distortions().copy(),
                   size=self.get_size(),
                   rvec=self.get_rotation().copy(),
                   tvec=self.get_translation().copy(),
                   name=self.get_name(),
                   extra_dist=self.extra_dist)

class FisheyeCamera(Camera):
    def __init__(self,
                 matrix=np.eye(3),
                 dist=np.zeros(4),
                 size=None,
                 rvec=np.zeros(3),
                 tvec=np.zeros(3),
                 name=None,
                 extra_dist=False):
        self.set_camera_matrix(matrix)
        self.set_distortions(dist)
        self.set_size(size)
        self.set_rotation(rvec)
        self.set_translation(tvec)
        self.set_name(name)
        self.extra_dist = extra_dist

    def from_dict(d):
        cam = FisheyeCamera()
        cam.load_dict(d)
        return cam

    def get_dict(self):
        d = super().get_dict()
        d['fisheye'] = True
        return d

    def distort_points(self, points):
        shape = points.shape
        points = points.reshape(-1, 1, 2)
        new_points = np.dstack([points, np.ones((points.shape[0], 1, 1))])
        out, _ = cv2.fisheye.projectPoints(new_points,
                                           np.zeros(3), np.zeros(3),
                                           self.matrix.astype('float64'),
                                           self.dist.astype('float64'))
        return out.reshape(shape)

    def undistort_points(self, points):
        shape = points.shape
        points = points.reshape(-1, 1, 2)
        out = cv2.fisheye.undistortPoints(points.astype('float64'),
                                          self.matrix.astype('float64'),
                                          self.dist.astype('float64'))
        return out.reshape(shape)

    def project(self, points):
        points = points.reshape(-1, 1, 3)
        out, _ = cv2.fisheye.projectPoints(points,
                                           self.rvec, self.tvec,
                                           self.matrix.astype('float64'),
                                           self.dist.astype('float64'))
        return out

    def set_params(self, params):
        self.set_rotation(params[0:3])
        self.set_translation(params[3:6])
        self.set_focal_length(params[6])

        dist = np.zeros(4, dtype='float64')
        dist[0] = params[7]
        if self.extra_dist:
            dist[1] = params[8]
        # dist[2] = params[9]
        # dist[3] = params[10]
        self.set_distortions(dist)

    def get_params(self):
        params = np.zeros(8+self.extra_dist, dtype='float64')
        params[0:3] = self.get_rotation()
        params[3:6] = self.get_translation()
        params[6] = self.get_focal_length()
        dist = self.get_distortions()
        params[7] = dist[0]
        if self.extra_dist:
            params[8] = dist[1]
        # params[9] = dist[2]
        # params[10] = dist[3]
        return params

    def copy(self):
        return FisheyeCamera(
            matrix=self.get_camera_matrix().copy(),
            dist=self.get_distortions().copy(),
            size=self.get_size(),
            rvec=self.get_rotation().copy(),
            tvec=self.get_translation().copy(),
            name=self.get_name(),
            extra_dist=self.extra_dist)

class CameraGroup:
    def __init__(self, cameras, metadata={}):
        self.cameras = cameras
        self.metadata = metadata

    def subset_cameras(self, indices):
        cams = [self.cameras[ix].copy() for ix in indices]
        return CameraGroup(cams, self.metadata)

    def subset_cameras_names(self, names):
        cur_names = self.get_names()
        cur_names_dict = dict(zip(cur_names, range(len(cur_names))))
        indices = []
        for name in names:
            if name not in cur_names_dict:
                raise IndexError(
                    "name {} not part of camera names: {}".format(
                        name, cur_names
                    ))
            indices.append(cur_names_dict[name])
        return self.subset_cameras(indices)

    def project(self, points):
        """Given an Nx3 array of points, this returns an CxNx2 array of 2D points,
        where C is the number of cameras"""
        points = points.reshape(-1, 1, 3)
        n_points = points.shape[0]
        n_cams = len(self.cameras)

        out = np.empty((n_cams, n_points, 2), dtype='float64')
        for cnum, cam in enumerate(self.cameras):
            out[cnum] = cam.project(points).reshape(n_points, 2)

        return out

    def triangulate(self, points, undistort=True, progress=False):
        """Given an CxNx2 array, this returns an Nx3 array of points,
        where N is the number of points and C is the number of cameras"""

        assert points.shape[0] == len(self.cameras), \
            "Invalid points shape, first dim should be equal to" \
            " number of cameras ({}), but shape is {}".format(
                len(self.cameras), points.shape
            )

        one_point = False
        if len(points.shape) == 2:
            points = points.reshape(-1, 1, 2)
            one_point = True

        if undistort:
            new_points = np.empty(points.shape)
            for cnum, cam in enumerate(self.cameras):
                # must copy in order to satisfy opencv underneath
                sub = np.copy(points[cnum])
                new_points[cnum] = cam.undistort_points(sub)
            points = new_points

        n_cams, n_points, _ = points.shape

        out = np.empty((n_points, 3))
        out[:] = np.nan

        cam_mats = np.array([cam.get_extrinsics_mat() for cam in self.cameras])

        if progress:
            iterator = trange(n_points, ncols=70)
        else:
            iterator = range(n_points)

        for ip in iterator:
            subp = points[:, ip, :]
            good = ~np.isnan(subp[:, 0])
            if np.sum(good) >= 2:
                out[ip] = triangulate_simple(subp[good], cam_mats[good])

        if one_point:
            out = out[0]

        return out

    def triangulate_possible(self, points, undistort=True,
                             min_cams=2, progress=False, threshold=0.5):
        """Given an CxNxPx2 array, this returns an Nx3 array of points
        by triangulating all possible points and picking the ones with
        best reprojection error
        where:
        C: number of cameras
        N: number of points
        P: number of possible options per point
        """

        assert points.shape[0] == len(self.cameras), \
            "Invalid points shape, first dim should be equal to" \
            " number of cameras ({}), but shape is {}".format(
                len(self.cameras), points.shape
            )

        n_cams, n_points, n_possible, _ = points.shape

        cam_nums, point_nums, possible_nums = np.where(
            ~np.isnan(points[:, :, :, 0]))

        all_iters = defaultdict(dict)

        for cam_num, point_num, possible_num in zip(cam_nums, point_nums,
                                                    possible_nums):
            if cam_num not in all_iters[point_num]:
                all_iters[point_num][cam_num] = []
            all_iters[point_num][cam_num].append((cam_num, possible_num))

        for point_num in all_iters.keys():
            for cam_num in all_iters[point_num].keys():
                all_iters[point_num][cam_num].append(None)

        out = np.full((n_points, 3), np.nan, dtype='float64')
        picked_vals = np.zeros((n_cams, n_points, n_possible), dtype='bool')
        errors = np.zeros(n_points, dtype='float64')
        points_2d = np.full((n_cams, n_points, 2), np.nan, dtype='float64')

        if progress:
            iterator = trange(n_points, ncols=70)
        else:
            iterator = range(n_points)

        for point_ix in iterator:
            best_point = None
            best_error = 200

            n_cams_max = len(all_iters[point_ix])

            for picked in itertools.product(*all_iters[point_ix].values()):
                picked = [p for p in picked if p is not None]
                if len(picked) < min_cams and len(picked) != n_cams_max:
                    continue

                cnums = [p[0] for p in picked]
                xnums = [p[1] for p in picked]

                pts = points[cnums, point_ix, xnums]
                cc = self.subset_cameras(cnums)

                p3d = cc.triangulate(pts, undistort=undistort)
                err = cc.reprojection_error(p3d, pts, mean=True)

                if err < best_error:
                    best_point = {
                        'error': err,
                        'point': p3d[:3],
                        'points': pts,
                        'picked': picked,
                        'joint_ix': point_ix
                    }
                    best_error = err
                    if best_error < threshold:
                        break

            if best_point is not None:
                out[point_ix] = best_point['point']
                picked = best_point['picked']
                cnums = [p[0] for p in picked]
                xnums = [p[1] for p in picked]
                picked_vals[cnums, point_ix, xnums] = True
                errors[point_ix] = best_point['error']
                points_2d[cnums, point_ix] = best_point['points']

        return out, picked_vals, points_2d, errors

    def triangulate_ransac(self, points, undistort=True, min_cams=2, progress=False):
        """Given an CxNx2 array, this returns an Nx3 array of points,
        where N is the number of points and C is the number of cameras"""

        assert points.shape[0] == len(self.cameras), \
            "Invalid points shape, first dim should be equal to" \
            " number of cameras ({}), but shape is {}".format(
                len(self.cameras), points.shape
            )

        n_cams, n_points, _ = points.shape

        points_ransac = points.reshape(n_cams, n_points, 1, 2)

        return self.triangulate_possible(points_ransac,
                                         undistort=undistort,
                                         min_cams=min_cams,
                                         progress=progress)


    @jit(parallel=True, forceobj=True)
    def reprojection_error(self, p3ds, p2ds, mean=False):
        """Given an Nx3 array of 3D points and an CxNx2 array of 2D points,
        where N is the number of points and C is the number of cameras,
        this returns an CxNx2 array of errors.
        Optionally mean=True, this averages the errors and returns array of length N of errors"""

        one_point = False
        if len(p3ds.shape) == 1 and len(p2ds.shape) == 2:
            p3ds = p3ds.reshape(1, 3)
            p2ds = p2ds.reshape(-1, 1, 2)
            one_point = True

        n_cams, n_points, _ = p2ds.shape
        assert p3ds.shape == (n_points, 3), \
            "shapes of 2D and 3D points are not consistent: " \
            "2D={}, 3D={}".format(p2ds.shape, p3ds.shape)

        errors = np.empty((n_cams, n_points, 2))

        for cnum, cam in enumerate(self.cameras):
            errors[cnum] = cam.reprojection_error(p3ds, p2ds[cnum])

        if mean:
            errors_norm = np.linalg.norm(errors, axis=2)
            good = ~np.isnan(errors_norm)
            errors_norm[~good] = 0
            denom = np.sum(good, axis=0).astype('float64')
            denom[denom < 1.5] = np.nan
            errors = np.sum(errors_norm, axis=0) / denom

        if one_point:
            if mean:
                errors = float(errors[0])
            else:
                errors = errors.reshape(-1, 2)

        return errors


    def bundle_adjust_iter(self, p2ds, extra=None,
                           n_iters=10, start_mu=15, end_mu=1,
                           max_nfev=200, ftol=1e-4,
                           n_samp_iter=100, n_samp_full=1000,
                           error_threshold=0.3,
                           verbose=False):
        """Given an CxNx2 array of 2D points,
        where N is the number of points and C is the number of cameras,
        this performs iterative bundle adjustsment to fine-tune the parameters of the cameras.
        That is, it performs bundle adjustment multiple times, adjusting the weights given to points
        to reduce the influence of outliers.
        This is inspired by the algorithm for Fast Global Registration by Zhou, Park, and Koltun
        """
        error_list = []
        
        assert p2ds.shape[0] == len(self.cameras), \
            "Invalid points shape, first dim should be equal to" \
            " number of cameras ({}), but shape is {}".format(
                len(self.cameras), p2ds.shape
            )

        p2ds_full = p2ds
        extra_full = extra

        p2ds, extra = resample_points(p2ds_full, extra_full,
                                      n_samp=n_samp_full)
        error = self.average_error(p2ds, median=True)

        if verbose:
            print('error: ', error)

        mus = np.exp(np.linspace(np.log(start_mu), np.log(end_mu), num=n_iters))

        if verbose:
            print('n_samples: {}'.format(n_samp_iter))

        for i in range(n_iters):
            p2ds, extra = resample_points(p2ds_full, extra_full,
                                          n_samp=n_samp_full)
            p3ds = self.triangulate(p2ds)
            errors_full = self.reprojection_error(p3ds, p2ds, mean=False)
            errors_norm = self.reprojection_error(p3ds, p2ds, mean=True)

            error_dict = get_error_dict(errors_full)
            max_error = 0
            min_error = 0
            for k, v in error_dict.items():
                num, percents = v
                max_error = max(percents[-1], max_error)
                min_error = max(percents[0], min_error)
            mu = max(min(max_error, mus[i]), min_error)

            good = errors_norm < mu
            extra_good = subset_extra(extra, good)
            p2ds_samp, extra_samp = resample_points(
                p2ds[:, good], extra_good, n_samp=n_samp_iter)

            error = np.median(errors_norm)
            error_list.append(error)

            if error < error_threshold:
                break

            if verbose:
                pprint(error_dict)
                print('error: {:.2f}, mu: {:.1f}, ratio: {:.3f}'.format(error, mu, np.mean(good)))
                print(f'previous error scores: {error_list}')

            self.bundle_adjust(p2ds_samp, extra_samp,
                               loss='linear', ftol=ftol,
                               max_nfev=max_nfev,
                               verbose=verbose)


        p2ds, extra = resample_points(p2ds_full, extra_full,
                                      n_samp=n_samp_full)
        p3ds = self.triangulate(p2ds)
        errors_full = self.reprojection_error(p3ds, p2ds, mean=False)
        errors_norm = self.reprojection_error(p3ds, p2ds, mean=True)
        error_dict = get_error_dict(errors_full)
        if verbose:
            pprint(error_dict)

        max_error = 0
        min_error = 0
        for k, v in error_dict.items():
            num, percents = v
            max_error = max(percents[-1], max_error)
            min_error = max(percents[0], min_error)
        mu = max(max(max_error, end_mu), min_error)

        good = errors_norm < mu
        extra_good = subset_extra(extra, good)
        self.bundle_adjust(p2ds[:, good], extra_good,
                           loss='linear',
                           ftol=ftol, max_nfev=max(200, max_nfev),
                           verbose=verbose)

        error = self.average_error(p2ds, median=True)

        p3ds = self.triangulate(p2ds)
        errors_full = self.reprojection_error(p3ds, p2ds, mean=False)
        error_dict = get_error_dict(errors_full)
        if verbose:
            pprint(error_dict)

        if verbose:
            print('error: ', error)

        return error

    def bundle_adjust(self, p2ds, extra=None,
                      loss='linear',
                      threshold=50,
                      ftol=1e-4,
                      max_nfev=1000,
                      weights=None,
                      start_params=None,
                      verbose=True):
        """Given an CxNx2 array of 2D points,
        where N is the number of points and C is the number of cameras,
        this performs bundle adjustsment to fine-tune the parameters of the cameras"""

        assert p2ds.shape[0] == len(self.cameras), \
            "Invalid points shape, first dim should be equal to" \
            " number of cameras ({}), but shape is {}".format(
                len(self.cameras), p2ds.shape
            )

        if extra is not None:
            extra['ids_map'] = remap_ids(extra['ids'])

        x0, n_cam_params = self._initialize_params_bundle(p2ds, extra)

        if start_params is not None:
            x0 = start_params
            n_cam_params = len(self.cameras[0].get_params())

        error_fun = self._error_fun_bundle

        jac_sparse = self._jac_sparsity_bundle(p2ds, n_cam_params, extra)

        f_scale = threshold
        opt = optimize.least_squares(error_fun,
                                     x0,
                                     jac_sparsity=jac_sparse,
                                     f_scale=f_scale,
                                     x_scale='jac',
                                     loss=loss,
                                     ftol=ftol,
                                     method='trf',
                                     tr_solver='lsmr',
                                     verbose=2 * verbose,
                                     max_nfev=max_nfev,
                                     args=(p2ds, n_cam_params, extra))
        best_params = opt.x

        for i, cam in enumerate(self.cameras):
            a = i * n_cam_params
            b = (i + 1) * n_cam_params
            cam.set_params(best_params[a:b])

        error = self.average_error(p2ds)
        return error

    @jit(parallel=True, forceobj=True)
    def _error_fun_bundle(self, params, p2ds, n_cam_params, extra):
        """Error function for bundle adjustment"""
        good = ~np.isnan(p2ds)
        n_cams = len(self.cameras)

        for i in range(n_cams):
            cam = self.cameras[i]
            a = i * n_cam_params
            b = (i + 1) * n_cam_params
            cam.set_params(params[a:b])

        n_cams = len(self.cameras)
        sub = n_cam_params * n_cams
        n3d = p2ds.shape[1] * 3
        p3ds_test = params[sub:sub+n3d].reshape(-1, 3)
        errors = self.reprojection_error(p3ds_test, p2ds)
        errors_reproj = errors[good]

        if extra is not None:
            ids = extra['ids_map']
            objp = extra['objp']
            min_scale = np.min(objp[objp > 0])
            n_boards = int(np.max(ids)) + 1
            a = sub+n3d
            rvecs = params[a:a+n_boards*3].reshape(-1, 3)
            tvecs = params[a+n_boards*3:a+n_boards*6].reshape(-1, 3)
            expected = transform_points(objp, rvecs[ids], tvecs[ids])
            errors_obj = 2 * (p3ds_test - expected).ravel() / min_scale
        else:
            errors_obj = np.array([])

        return np.hstack([errors_reproj, errors_obj])


    def _jac_sparsity_bundle(self, p2ds, n_cam_params, extra):
        """Given an CxNx2 array of 2D points,
        where N is the number of points and C is the number of cameras,
        compute the sparsity structure of the jacobian for bundle adjustment"""

        point_indices = np.zeros(p2ds.shape, dtype='int32')
        cam_indices = np.zeros(p2ds.shape, dtype='int32')

        for i in range(p2ds.shape[1]):
            point_indices[:, i] = i

        for j in range(p2ds.shape[0]):
            cam_indices[j] = j

        good = ~np.isnan(p2ds)

        if extra is not None:
            ids = extra['ids_map']
            n_boards = int(np.max(ids)) + 1
            total_board_params = n_boards * (3 + 3) # rvecs + tvecs
        else:
            n_boards = 0
            total_board_params = 0

        n_cams = p2ds.shape[0]
        n_points = p2ds.shape[1]
        total_params_reproj = n_cams * n_cam_params + n_points * 3
        n_params = total_params_reproj + total_board_params

        n_good_values = np.sum(good)
        if extra is not None:
            n_errors = n_good_values + n_points * 3
        else:
            n_errors = n_good_values

        A_sparse = dok_matrix((n_errors, n_params), dtype='int16')

        cam_indices_good = cam_indices[good]
        point_indices_good = point_indices[good]

        # -- reprojection error --
        ix = np.arange(n_good_values)

        ## update camera params based on point error
        for i in range(n_cam_params):
            A_sparse[ix, cam_indices_good * n_cam_params + i] = 1

        ## update point position based on point error
        for i in range(3):
            A_sparse[ix, n_cams * n_cam_params + point_indices_good * 3 + i] = 1

        # -- match for the object points--
        if extra is not None:
            point_ix = np.arange(n_points)

            ## update all the camera parameters
            # A_sparse[n_good_values:n_good_values+n_points*3,
            #          0:n_cams*n_cam_params] = 1

            ## update board rotation and translation based on error from expected
            for i in range(3):
                for j in range(3):
                    A_sparse[n_good_values + point_ix*3 + i,
                             total_params_reproj + ids*3 + j] = 1
                    A_sparse[n_good_values + point_ix*3 + i,
                             total_params_reproj + n_boards*3 + ids*3 + j] = 1


            ## update point position based on error from expected
            for i in range(3):
                A_sparse[n_good_values + point_ix*3 + i,
                         n_cams*n_cam_params + point_ix*3 + i] = 1


        return A_sparse

    def _initialize_params_bundle(self, p2ds, extra):
        """Given an CxNx2 array of 2D points,
        where N is the number of points and C is the number of cameras,
        initializes the parameters for bundle adjustment"""

        cam_params = np.hstack([cam.get_params() for cam in self.cameras])
        n_cam_params = len(cam_params) // len(self.cameras)

        total_cam_params = len(cam_params)

        n_cams, n_points, _ = p2ds.shape
        assert n_cams == len(self.cameras), \
            "number of cameras in CameraGroup does not " \
            "match number of cameras in 2D points given"

        p3ds = self.triangulate(p2ds)

        if extra is not None:
            ids = extra['ids_map']
            n_boards = int(np.max(ids[~np.isnan(ids)])) + 1
            total_board_params = n_boards * (3 + 3) # rvecs + tvecs

            # initialize to 0
            rvecs = np.zeros((n_boards, 3), dtype='float64')
            tvecs = np.zeros((n_boards, 3), dtype='float64')

            if 'rvecs' in extra and 'tvecs' in extra:
                rvecs_all = extra['rvecs']
                tvecs_all = extra['tvecs']
                for board_num in range(n_boards):
                    point_id = np.where(ids == board_num)[0][0]
                    cam_ids_possible = np.where(~np.isnan(p2ds[:, point_id, 0]))[0]
                    cam_id = np.random.choice(cam_ids_possible)
                    M_cam = self.cameras[cam_id].get_extrinsics_mat()
                    M_board_cam = make_M(rvecs_all[cam_id, point_id],
                                         tvecs_all[cam_id, point_id])
                    M_board = np.matmul(np.linalg.inv(M_cam), M_board_cam)
                    rvec, tvec = get_rtvec(M_board)
                    rvecs[board_num] = rvec
                    tvecs[board_num] = tvec


        else:
            total_board_params = 0

        x0 = np.zeros(total_cam_params + p3ds.size + total_board_params)
        x0[:total_cam_params] = cam_params
        x0[total_cam_params:total_cam_params+p3ds.size] = p3ds.ravel()

        if extra is not None:
            start_board = total_cam_params+p3ds.size
            x0[start_board:start_board + n_boards*3] = rvecs.ravel()
            x0[start_board + n_boards*3:start_board + n_boards*6] = \
                tvecs.ravel()

        return x0, n_cam_params

    def optim_points(self, points, p3ds,
                     constraints=[],
                     constraints_weak=[],
                     scale_smooth=4,
                     scale_length=2, scale_length_weak=0.5,
                     reproj_error_threshold=15, reproj_loss='soft_l1',
                     n_deriv_smooth=1, scores=None, verbose=False):
        """
        Take in an array of 2D points of shape CxNxJx2,
        an array of 3D points of shape NxJx3,
        and an array of constraints of shape Kx2, where
        C: number of camera
        N: number of frames
        J: number of joints
        K: number of constraints

        This function creates an optimized array of 3D points of shape NxJx3.

        Example constraints:
        constraints = [[0, 1], [1, 2], [2, 3]]
        (meaning that lengths of segments 0->1, 1->2, 2->3 are all constant)

        """
        assert points.shape[0] == len(self.cameras), \
            "Invalid points shape, first dim should be equal to" \
            " number of cameras ({}), but shape is {}".format(
                len(self.cameras), points.shape
            )

        n_cams, n_frames, n_joints, _ = points.shape
        constraints = np.array(constraints)
        constraints_weak = np.array(constraints_weak)

        p3ds_intp = np.apply_along_axis(interpolate_data, 0, p3ds)

        p3ds_med = np.apply_along_axis(medfilt_data, 0, p3ds_intp, size=7)

        default_smooth = 1.0/np.mean(np.abs(np.diff(p3ds_med, axis=0)))
        scale_smooth_full = scale_smooth * default_smooth

        t1 = time.time()

        x0 = self._initialize_params_triangulation(
            p3ds_intp, constraints, constraints_weak)

        x0[~np.isfinite(x0)] = 0

        jac = self._jac_sparsity_triangulation(
            points, constraints, constraints_weak, n_deriv_smooth)

        opt2 = optimize.least_squares(self._error_fun_triangulation,
                                      x0=x0, jac_sparsity=jac,
                                      loss='linear',
                                      ftol=1e-3,
                                      verbose=2*verbose,
                                      args=(points,
                                            constraints,
                                            constraints_weak,
                                            scores,
                                            scale_smooth_full,
                                            scale_length,
                                            scale_length_weak,
                                            reproj_error_threshold,
                                            reproj_loss,
                                            n_deriv_smooth))

        p3ds_new2 = opt2.x[:p3ds.size].reshape(p3ds.shape)

        t2 = time.time()

        if verbose:
            print('optimization took {:.2f} seconds'.format(t2 - t1))

        return p3ds_new2


    def optim_points_possible(self, points, p3ds,
                              constraints=[],
                              constraints_weak=[],
                              scale_smooth=4,
                              scale_length=2, scale_length_weak=0.5,
                              reproj_error_threshold=15, reproj_loss='soft_l1',
                              n_deriv_smooth=1, scores=None, verbose=False):
        """
        Take in an array of 2D points of shape CxNxJxPx2,
        an array of 3D points of shape NxJx3,
        and an array of constraints of shape Kx2, where
        C: number of camera
        N: number of frames
        J: number of joints
        P: number of possible options per point
        K: number of constraints

        This function creates an optimized array of 3D points of shape NxJx3.

        Example constraints:
        constraints = [[0, 1], [1, 2], [2, 3]]
        (meaning that lengths of segments 0->1, 1->2, 2->3 are all constant)

        """
        assert points.shape[0] == len(self.cameras), \
            "Invalid points shape, first dim should be equal to" \
            " number of cameras ({}), but shape is {}".format(
                len(self.cameras), points.shape
            )

        n_cams, n_frames, n_joints, n_possible, _ = points.shape
        constraints = np.array(constraints)
        constraints_weak = np.array(constraints_weak)

        p3ds_intp = np.apply_along_axis(interpolate_data, 0, p3ds)

        p3ds_med = np.apply_along_axis(medfilt_data, 0, p3ds_intp, size=7)

        default_smooth = 1.0/np.mean(np.abs(np.diff(p3ds_med, axis=0)))
        scale_smooth_full = scale_smooth * default_smooth

        t1 = time.time()

        x0 = self._initialize_params_triangulation_possible(
            p3ds_intp, points, constraints=constraints, constraints_weak=constraints_weak)

        print('getting jacobian...')
        jac = self._jac_sparsity_triangulation_possible(
            points,
            constraints=constraints,
            constraints_weak=constraints_weak,
            n_deriv_smooth=n_deriv_smooth)

        beta = 5

        print('starting optimization...')
        opt2 = optimize.least_squares(self._error_fun_triangulation_possible,
                                      x0=x0, jac_sparsity=jac,
                                      loss='linear',
                                      ftol=1e-3,
                                      verbose=2*verbose,
                                      args=(points,
                                            beta,
                                            constraints,
                                            constraints_weak,
                                            scores,
                                            scale_smooth_full,
                                            scale_length,
                                            scale_length_weak,
                                            reproj_error_threshold,
                                            reproj_loss,
                                            n_deriv_smooth))
        params = opt2.x

        p3ds_new2 = params[:p3ds.size].reshape(p3ds.shape)

        bad = np.isnan(points[:, :, :, :, 0])
        all_bad = np.all(bad, axis=3)

        n_params_norm = p3ds.size + len(constraints) + len(constraints_weak)

        alphas = np.zeros((n_cams, n_frames, n_joints, n_possible), dtype='float64')
        alphas[~bad] = params[n_params_norm:]

        alphas_exp = np.exp(beta * alphas)
        alphas_exp[bad] = 0
        alphas_sum = np.sum(alphas_exp, axis=3)
        alphas_sum[all_bad] = 1
        alphas_norm = alphas_exp / alphas_sum[:, :, :, None]
        alphas_norm[bad] = np.nan

        t2 = time.time()

        if verbose:
            print('optimization took {:.2f} seconds'.format(t2 - t1))

        return p3ds_new2, alphas_norm


    def triangulate_optim(self, points, init_ransac=False, init_progress=False,
                          **kwargs):
        """
        Take in an array of 2D points of shape CxNxJx2, and an array of constraints of shape Kx2, where
        C: number of camera
        N: number of frames
        J: number of joints
        K: number of constraints

        This function creates an optimized array of 3D points of shape NxJx3.

        Example constraints:
        constraints = [[0, 1], [1, 2], [2, 3]]
        (meaning that lengths of segments 0->1, 1->2, 2->3 are all constant)

        """

        assert points.shape[0] == len(self.cameras), \
            "Invalid points shape, first dim should be equal to" \
            " number of cameras ({}), but shape is {}".format(
                len(self.cameras), points.shape
            )

        n_cams, n_frames, n_joints, _ = points.shape
        # constraints = np.array(constraints)
        # constraints_weak = np.array(constraints_weak)

        points_shaped = points.reshape(n_cams, n_frames*n_joints, 2)
        if init_ransac:
            p3ds, picked, p2ds, errors = self.triangulate_ransac(points_shaped, progress=init_progress)
            points = p2ds.reshape(points.shape)
        else:
            p3ds = self.triangulate(points_shaped, progress=init_progress)
        p3ds = p3ds.reshape((n_frames, n_joints, 3))

        c = np.isfinite(p3ds[:, :, 0])
        if np.sum(c) < 20:
            print("warning: not enough 3D points to run optimization")
            return p3ds

        return self.optim_points(points, p3ds, **kwargs)



    @jit(forceobj=True, parallel=True)
    def _error_fun_triangulation(self, params, p2ds,
                                 constraints=[],
                                 constraints_weak=[],
                                 scores=None,
                                 scale_smooth=10000,
                                 scale_length=1,
                                 scale_length_weak=0.2,
                                 reproj_error_threshold=100,
                                 reproj_loss='soft_l1',
                                 n_deriv_smooth=1):
        n_cams, n_frames, n_joints, _ = p2ds.shape

        n_3d = n_frames*n_joints*3
        n_constraints = len(constraints)
        n_constraints_weak = len(constraints_weak)

        # load params
        p3ds = params[:n_3d].reshape((n_frames, n_joints, 3))
        joint_lengths = np.array(params[n_3d:n_3d+n_constraints])
        joint_lengths_weak = np.array(params[n_3d+n_constraints:])

        # reprojection errors
        p3ds_flat = p3ds.reshape(-1, 3)
        p2ds_flat = p2ds.reshape((n_cams, -1, 2))
        errors = self.reprojection_error(p3ds_flat, p2ds_flat)
        if scores is not None:
            scores_flat = scores.reshape((n_cams, -1))
            errors = errors * scores_flat[:, :, None]
        errors_reproj = errors[~np.isnan(p2ds_flat)]

        rp = reproj_error_threshold
        errors_reproj = np.abs(errors_reproj)
        if reproj_loss == 'huber':
            bad = errors_reproj > rp
            errors_reproj[bad] = rp*(2*np.sqrt(errors_reproj[bad]/rp) - 1)
        elif reproj_loss == 'linear':
            pass
        elif reproj_loss == 'soft_l1':
            errors_reproj = rp*2*(np.sqrt(1+errors_reproj/rp)-1)

        # temporal constraint
        errors_smooth = np.diff(p3ds, n=n_deriv_smooth, axis=0).ravel() * scale_smooth

        # joint length constraint
        errors_lengths = np.empty((n_constraints, n_frames), dtype='float64')
        for cix, (a, b) in enumerate(constraints):
            lengths = np.linalg.norm(p3ds[:, a] - p3ds[:, b], axis=1)
            expected = joint_lengths[cix]
            errors_lengths[cix] = 100*(lengths - expected)/expected
        errors_lengths = errors_lengths.ravel() * scale_l0ength

        errors_lengths_weak = np.empty((n_constraints_weak, n_frames), dtype='float64')
        for cix, (a, b) in enumerate(constraints_weak):
            lengths = np.linalg.norm(p3ds[:, a] - p3ds[:, b], axis=1)
            expected = joint_lengths_weak[cix]
            errors_lengths_weak[cix] = 100*(lengths - expected)/expected
        errors_lengths_weak = errors_lengths_weak.ravel() * scale_length_weak

        return np.hstack([errors_reproj, errors_smooth,
                          errors_lengths, errors_lengths_weak])

    def _error_fun_triangulation_possible(self, params, p2ds,
                                          beta=2,
                                          constraints=[],
                                          constraints_weak=[],
                                          *args):
        # extract alphas from end of params
        # soft argmax for picking the appropriate points from p2ds
        # pass the points to error_fun_triangulate_possible for residuals
        # add errors to keep the alphas in check
        # return all the errors

        n_cams, n_frames, n_joints, n_possible, _ = p2ds.shape

        n_3d = n_frames*n_joints*3
        n_constraints = len(constraints)
        n_constraints_weak = len(constraints_weak)
        n_params_norm = n_3d+n_constraints+n_constraints_weak

        # load params
        bad = np.isnan(p2ds[:, :, :, :, 0])
        all_bad = np.all(bad, axis=3)

        alphas = np.zeros((n_cams, n_frames, n_joints, n_possible), dtype='float64')
        alphas[~bad] = params[n_params_norm:]
        params_rest = np.array(params[:n_params_norm])

        # get normalized alphas
        alphas_exp = np.exp(beta * alphas)
        alphas_exp[bad] = 0
        alphas_sum = np.sum(alphas_exp, axis=3)
        alphas_sum[all_bad] = 1
        alphas_norm = alphas_exp / alphas_sum[:, :, :, None]

        # extract the 2D points using soft argmax
        p2ds_test = np.copy(p2ds)
        p2ds_test[bad] = 0
        p2ds_adj = np.sum(alphas_norm[:, :, :, :, None] * p2ds_test, axis=3)
        p2ds_adj[all_bad] = np.nan

        errors = self._error_fun_triangulation(params_rest, p2ds_adj,
                                               constraints, constraints_weak, *args)

        alphas_test = alphas_norm[~all_bad]
        errors_alphas = (1 - np.std(alphas_test, axis=1)) * 10

        return np.hstack([errors, errors_alphas])


    def    _initialize_params_triangulation(self, p3ds,
                                         constraints=[],
                                         constraints_weak=[]):
        joint_lengths = np.empty(len(constraints), dtype='float64')
        joint_lengths_weak = np.empty(len(constraints_weak), dtype='float64')

        for cix, (a, b) in enumerate(constraints):
            lengths = np.linalg.norm(p3ds[:, a] - p3ds[:, b], axis=1)
            joint_lengths[cix] = np.median(lengths)


        for cix, (a, b) in enumerate(constraints_weak):
            lengths = np.linalg.norm(p3ds[:, a] - p3ds[:, b], axis=1)
            joint_lengths_weak[cix] = np.median(lengths)

        all_lengths = np.hstack([joint_lengths, joint_lengths_weak])
        med = np.median(all_lengths)
        if med == 0:
            med = 1e-3

        mad = np.median(np.abs(all_lengths - med))

        joint_lengths[joint_lengths == 0] = med
        joint_lengths_weak[joint_lengths_weak == 0] = med
        joint_lengths[joint_lengths > med+mad*5] = med
        joint_lengths_weak[joint_lengths_weak > med+mad*5] = med

        return np.hstack([p3ds.ravel(), joint_lengths, joint_lengths_weak])

    def _initialize_params_triangulation_possible(self, p3ds, p2ds, **kwargs):
        # initialize params using above function
        # initialize alphas to 1 for first one and 0 for other possible

        n_cams, n_frames, n_joints, n_possible, _ = p2ds.shape
        good = ~np.isnan(p2ds[:, :, :, :, 0])

        alphas = np.zeros((n_cams, n_frames, n_joints, n_possible), dtype='float64')
        alphas[:, :, :, 0] = 0

        params = self._initialize_params_triangulation(p3ds, **kwargs)
        params_full = np.hstack([params, alphas[good]])

        return params_full

    def _jac_sparsity_triangulation(self, p2ds,
                                    constraints=[],
                                    constraints_weak=[],
                                    n_deriv_smooth=1):
        n_cams, n_frames, n_joints, _ = p2ds.shape
        n_constraints = len(constraints)
        n_constraints_weak = len(constraints_weak)

        p2ds_flat = p2ds.reshape((n_cams, -1, 2))

        point_indices = np.zeros(p2ds_flat.shape, dtype='int32')
        for i in range(p2ds_flat.shape[1]):
            point_indices[:, i] = i

        point_indices_3d = np.arange(n_frames*n_joints)\
                             .reshape((n_frames, n_joints))

        good = ~np.isnan(p2ds_flat)
        n_errors_reproj = np.sum(good)
        n_errors_smooth = (n_frames-n_deriv_smooth) * n_joints * 3
        n_errors_lengths = n_constraints * n_frames
        n_errors_lengths_weak = n_constraints_weak * n_frames

        n_errors = n_errors_reproj + n_errors_smooth + \
            n_errors_lengths + n_errors_lengths_weak

        n_3d = n_frames*n_joints*3
        n_params = n_3d + n_constraints + n_constraints_weak

        point_indices_good = point_indices[good]

        A_sparse = dok_matrix((n_errors, n_params), dtype='int16')

        # constraints for reprojection errors
        ix_reproj = np.arange(n_errors_reproj)
        for k in range(3):
            A_sparse[ix_reproj, point_indices_good * 3 + k] = 1

        # sparse constraints for smoothness in time
        frames = np.arange(n_frames-n_deriv_smooth)
        for j in range(n_joints):
            for n in range(n_deriv_smooth+1):
                pa = point_indices_3d[frames, j]
                pb = point_indices_3d[frames+n, j]
                for k in range(3):
                    A_sparse[n_errors_reproj + pa*3 + k, pb*3 + k] = 1

        ## -- strong constraints --
        # joint lengths should change with joint lengths errors
        start = n_errors_reproj + n_errors_smooth
        frames = np.arange(n_frames)
        for cix, (a, b) in enumerate(constraints):
            A_sparse[start + cix*n_frames + frames, n_3d+cix] = 1

        # points should change accordingly to match joint lengths too
        frames = np.arange(n_frames)
        for cix, (a, b) in enumerate(constraints):
            pa = point_indices_3d[frames, a]
            pb = point_indices_3d[frames, b]
            for k in range(3):
                A_sparse[start + cix*n_frames + frames, pa*3 + k] = 1
                A_sparse[start + cix*n_frames + frames, pb*3 + k] = 1

        ## -- weak constraints --
        # joint lengths should change with joint lengths errors
        start = n_errors_reproj + n_errors_smooth + n_errors_lengths
        frames = np.arange(n_frames)
        for cix, (a, b) in enumerate(constraints_weak):
            A_sparse[start + cix*n_frames + frames, n_3d + n_constraints + cix] = 1

        # points should change accordingly to match joint lengths too
        frames = np.arange(n_frames)
        for cix, (a, b) in enumerate(constraints_weak):
            pa = point_indices_3d[frames, a]
            pb = point_indices_3d[frames, b]
            for k in range(3):
                A_sparse[start + cix*n_frames + frames, pa*3 + k] = 1
                A_sparse[start + cix*n_frames + frames, pb*3 + k] = 1

        return A_sparse

    def _jac_sparsity_triangulation_possible(self, p2ds_full, **kwargs):
        # initialize sparse jacobian using above function
        # extend to include alphas from parameters
        ## TODO: this initialization is really slow for some reason

        n_cams, n_frames, n_joints, n_possible, _ = p2ds_full.shape
        good_full = ~np.isnan(p2ds_full[:, :, :, :, 0])
        any_good = np.any(good_full, axis=3)

        n_alphas = np.sum(good_full)
        n_errors_alphas = np.sum(any_good)

        p2ds = p2ds_full[:, :, :, 0]
        A_sparse = self._jac_sparsity_triangulation(p2ds, **kwargs)

        n_errors, n_params = A_sparse.shape

        B_sparse = dok_matrix((n_errors + n_errors_alphas, n_params + n_alphas), dtype='int16')
        for r, c in zip(*A_sparse.nonzero()):
            B_sparse[r, c] = A_sparse[r, c]

        point_indices_2d = np.arange(n_cams*n_frames*n_joints)\
                             .reshape(n_cams, n_frames, n_joints)
        point_indices_2d_rep = np.repeat(point_indices_2d[:, :, :, None], 2, axis=3)
        point_indices_2d_good = point_indices_2d_rep[~np.isnan(p2ds)]
        point_indices_good = point_indices_2d[any_good]

        alpha_indices = np.zeros((n_cams, n_frames, n_joints, n_possible), dtype='int64')
        for pnum in range(n_possible):
            alpha_indices[:, :, :, pnum] = point_indices_2d

        alpha_indices_good = alpha_indices[good_full]

        # alphas should change according to the reprojection error for each corresponding point
        point_indices_2d_good_find = defaultdict(list)
        for ix, p in enumerate(point_indices_2d_good):
            point_indices_2d_good_find[p].append(ix)

        for ix, alpha_index in enumerate(alpha_indices_good):
            B_sparse[point_indices_2d_good_find[alpha_index],
                     n_params + ix] = 1

        # alphas should change according to the alpha errors
        point_indices_good_find = dict()
        for ix, p in enumerate(point_indices_good):
            point_indices_good_find[p] = ix

        for ix, alpha_index in enumerate(alpha_indices_good):
            if alpha_index in point_indices_good_find:
                err_ix = n_errors + point_indices_good_find[alpha_index]
                B_sparse[err_ix, n_params + ix] = 1

        return B_sparse

    def copy(self):
        cameras = [cam.copy() for cam in self.cameras]
        metadata = copy(self.metadata)
        return CameraGroup(cameras, metadata)

    def set_rotations(self, rvecs):
        for cam, rvec in zip(self.cameras, rvecs):
            cam.set_rotation(rvec)

    def set_translations(self, tvecs):
        for cam, tvec in zip(self.cameras, tvecs):
            cam.set_translation(tvec)

    def get_rotations(self):
        rvecs = []
        for cam in self.cameras:
            rvec = cam.get_rotation()
            rvecs.append(rvec)
        return np.array(rvecs)

    def get_translations(self):
        tvecs = []
        for cam in self.cameras:
            tvec = cam.get_translation()
            tvecs.append(tvec)
        return np.array(tvecs)

    def get_names(self):
        return [cam.get_name() for cam in self.cameras]

    def set_names(self, names):
        for cam, name in zip(self.cameras, names):
            cam.set_name(name)

    def average_error(self, p2ds, median=False):
        p3ds = self.triangulate(p2ds)
        errors = self.reprojection_error(p3ds, p2ds, mean=True)
        if median:
            return np.median(errors)
        else:
            return np.mean(errors)

    def calibrate_rows(self, all_rows, board,
                       init_intrinsics=True, init_extrinsics=True, verbose=True,
                       **kwargs):
        assert len(all_rows) == len(self.cameras), \
            "Number of camera detections does not match number of cameras"

        for rows, camera in zip(all_rows, self.cameras):
            size = camera.get_size()

            assert size is not None, \
                "Camera with name {} has no specified frame size".format(camera.get_name())

            if init_intrinsics:
                objp, imgp = board.get_all_calibration_points(rows)
                mixed = [(o, i) for (o, i) in zip(objp, imgp) if len(o) >= 7]
                assert len(objp) != 0 and len(imgp) != 0, "No Charuco board points detected"
                objp, imgp = zip(*mixed)
                matrix = cv2.initCameraMatrix2D(objp, imgp, tuple(size))
                camera.set_camera_matrix(matrix)

        for i, (row, cam) in enumerate(zip(all_rows, self.cameras)):
            all_rows[i] = board.estimate_pose_rows(cam, row)

        charuco_frames = [f['framenum'][1] for f in all_rows[0]]
        merged = merge_rows(all_rows)
        imgp, extra = extract_points(merged, board, min_cameras=2)

        if init_extrinsics:
            rtvecs = extract_rtvecs(merged)
            if verbose:
                pprint(get_connections(rtvecs, self.get_names()))
            rvecs, tvecs = get_initial_extrinsics(rtvecs)
            self.set_rotations(rvecs)
            self.set_translations(tvecs)

        error = self.bundle_adjust_iter(imgp, extra, verbose=verbose, error_threshold=1)

        return error,merged,charuco_frames 

    def get_rows_videos(self, videos, board, verbose=True):
        all_rows = []

        for cix, (cam, cam_videos) in enumerate(zip(self.cameras, videos)):
            rows_cam = []
            for vnum, vidname in enumerate(cam_videos):
                if verbose: print(vidname)
                rows = board.detect_video(vidname, prefix=vnum, progress=verbose)
                if verbose: print("{} boards detected".format(len(rows)))
                rows_cam.extend(rows)
            all_rows.append(rows_cam)

        return all_rows

    def set_camera_sizes_videos(self, videos):
        for cix, (cam, cam_videos) in enumerate(zip(self.cameras, videos)):
            rows_cam = []
            for vnum, vidname in enumerate(cam_videos):
                params = get_video_params(vidname)
                size = (params['width'], params['height'])
                cam.set_size(size)

    def calibrate_videos(self, videos, board,
                         init_intrinsics=True, init_extrinsics=True, verbose=True,
                         **kwargs):
        """Takes as input a list of list of video filenames, one list of each camera.
        Also takes a board which specifies what should be detected in the videos"""

        all_rows = self.get_rows_videos(videos, board, verbose=verbose)
        if init_extrinsics:
            self.set_camera_sizes_videos(videos)

        error,merged,charuco_frames = self.calibrate_rows(all_rows, board,
                                    init_intrinsics=init_intrinsics,
                                    init_extrinsics=init_extrinsics,
                                    verbose=verbose, **kwargs)
        return error, merged, charuco_frames

    def get_dicts(self):
        out = []
        for cam in self.cameras:
            out.append(cam.get_dict())
        return out

    def from_dicts(arr):
        cameras = []
        for d in arr:
            if 'fisheye' in d and d['fisheye']:
                cam = FisheyeCamera.from_dict(d)
            else:
                cam = Camera.from_dict(d)
            cameras.append(cam)
        return CameraGroup(cameras)

    def from_names(names, fisheye=False):
        cameras = []
        for name in names:
            if fisheye:
                cam = FisheyeCamera(name=name)
            else:
                cam = Camera(name=name)
            cameras.append(cam)
        return CameraGroup(cameras)

    def load_dicts(self, arr):
        for cam, d in zip(self.cameras, arr):
            cam.load_dict(d)

    def dump(self, fname):
        dicts = self.get_dicts()
        names = ['cam_{}'.format(i) for i in range(len(dicts))]
        master_dict = dict(zip(names, dicts))
        master_dict['metadata'] = self.metadata
        with open(fname, 'w') as f:
            toml.dump(master_dict, f)

    def load(fname):
        master_dict = toml.load(fname)
        keys = sorted(master_dict.keys())
        items = [master_dict[k] for k in keys if k != 'metadata']
        cgroup = CameraGroup.from_dicts(items)
        if 'metadata' in master_dict:
            cgroup.metadata = master_dict['metadata']
        return cgroup

    def resize_cameras(self, scale):
        for cam in self.cameras:
            cam.resize_camera(scale)



import os
from pathlib import Path

BASE_FOLDER_NAME = "freemocap_data"

freemocap_data_path = Path().joinpath(Path.home(), BASE_FOLDER_NAME)


def create_home_data_directory():
    path_as_str = str(freemocap_data_path)
    if not os.path.exists(path_as_str):
        os.makedirs(path_as_str)


def create_path(*args):
    return Path().joinpath(*args)



from freemocap.webcam import startcamrecording, timesync, videotrim

from pathlib import Path
import time
import pickle
import pandas as pd
import numpy as np
from tkinter import Tk


def RecordCams(session,camInputs,parameterDictionary,rotationInputs):
    """ 
    Determines the number of cameras, assigns them IDs, and then starts a threaded cam recording process. Accesses the pickle file of timestamps
    saved during recording, makes a dataframe from it and saves it to a CSV file. Updates the session class at the end with numCam variable
    """  
    #Create RawVideos folder
    session.rawVidPath.mkdir(exist_ok = True)

    #%% Setting up recordings
    beginTime = time.time()
    session.beginTime = beginTime
    numCams = len(camInputs)  # number of cameras
    numCamRange = range(numCams)  # a range for the number of cameras that we have
    vidNames = []
    camIDs = []
    unix_camIDs = []
    for x in numCamRange:  # create names for each of the initial untrimmed videos
        singleCamID = "Cam{}".format(x + 1)
        camIDs.append(
            singleCamID
        )  # creates IDs for each camera based on the number of cameras entered
        
        unix_camID = singleCamID + '_unix_timestamps'
        unix_camIDs.append(unix_camID)

        singleVidName = "raw_cam{}.mp4".format(x + 1)
        vidNames.append(singleVidName)

    #%% Starting the thread recordings for each camera
    threads = []
    for n in numCamRange:  # starts recording video, opens threads for each camera
        camRecordings = startcamrecording.CamRecordingThread(
            session,
            camIDs[n],
            unix_camIDs[n],
            camInputs[n],
            vidNames[n],
            session.rawVidPath,
            beginTime,
            parameterDictionary,
        )
        camRecordings.start()

        threads.append(camRecordings)

    for camRecordings in threads:
        camRecordings.join()  # make sure that one thread ending doesn't immediately end all the others (before they can dump data in a pickle file)

    print("finished recordings")

    timeStampList = []
    unix_timeStampList = []

    for (
        e
    ) in (
        numCamRange
    ):  # open the saved pickle file for each camera, and add the timestamps to the dataList list
        with open(session.rawVidPath/camIDs[e], "rb") as f:
            camTimeList = pickle.load(f)
            timeStampList.append(camTimeList)
        with open(session.rawVidPath/unix_camIDs[e], "rb") as g:
            unix_camTimeList = pickle.load(g)
            unix_timeStampList.append(unix_camTimeList)

    timeDictionary = {}
    unix_timeDictionary = {}

    id_and_time = zip(camIDs, timeStampList)

    for cam, data in id_and_time:
        timeDictionary[cam] = np.array(
            data
        )  # create a dictionary that holds the timestamps for each camera
    df = pd.DataFrame.from_dict(
        timeDictionary, orient="index"
    )  # create a data frame from this dictionary
    timeStampData = df.transpose()
    csvName = session.sessionID + "_timestamps.csv" 
    csvPath = session.rawVidPath / csvName
    timeStampData.to_csv(csvPath)  # turn dataframe into a CSV

    id_and_unix_time = zip(unix_camIDs,unix_timeStampList)
    
    for cam_unix, data_unix in id_and_unix_time:
        unix_timeDictionary[cam_unix] = np.array(data_unix)
    df_unix = pd.DataFrame.from_dict(
    unix_timeDictionary, orient="index"
    )  # create a data frame from this dictionary
    unix_timeStampData = df_unix.transpose()
    unix_csvName = session.sessionID + "_unix_timestamps.csv"
    unix_csvPath = session.rawVidPath / unix_csvName
    unix_timeStampData.to_csv(unix_csvPath)

    session.numCams = numCams
    session.session_settings['recording_parameters'].update({'numCams':session.numCams})
    session.timeStampData = timeStampData
    session.camIDs = camIDs
    session.numCamRange = numCamRange
    session.vidNames = vidNames
    

def SyncCams(session, timeStampData,numCamRange,vidNames,camIDs):
    """ 
    Runs the time-syncing process. Accesses saved timestamps, runs the time-syncing GUI, and on user-permission, proceeds to create
    synced videos 
    """  
    session.syncedVidPath.mkdir(exist_ok = True)

    #start the timesync process
    frameTable,timeTable,unix_synced_timeTable,frameRate,resultsTable,plots = timesync.TimeSync(session,timeStampData,numCamRange,camIDs) 
    
    #this message shows you your percentages and asks if you would like to continue or not. shuts down the program if no
    root = Tk()
    proceed = timesync.proceedGUI(
        root, resultsTable, plots
    )  # create a GUI instance called proceed
    root.mainloop()

    if session.get_synced_unix_timestamps == True:
        unix_synced_timestamps_csvName = 'unix_synced_timestamps.csv'
        unix_synced_timestamps_csvPath = session.sessionPath/unix_synced_timestamps_csvName
        unix_synced_timeTable.to_csv(unix_synced_timestamps_csvPath)

    if proceed.proceed == True:
        print()
        print('Starting editing')
        videotrim.VideoTrim(session,vidNames,frameTable,session.parameterDictionary,session.rotationInputs,numCamRange)
        session.session_settings['recording_parameters'].update({'numFrames':session.numFrames})
        #videotrim.createCalibrationVideos(session,60,parameterDictionary)
        print('all done')




from rich import print
from rich.console import Console
from FMC_MultiCamera import FMC_MultiCamera
import socket
from pathlib import Path
from pathos.helpers import mp as pathos_mp_helper
import cv2
from copy import deepcopy
import numpy as np

from itertools import combinations

import matplotlib.pyplot as plt

aruco_dict = cv2.aruco.Dictionary_get(cv2.aruco.DICT_4X4_250)
charuco_length = 7
charuco_width = 5

board = cv2.aruco.CharucoBoard_create(charuco_length, charuco_width, 1, 0.8, aruco_dict)
num_charuco_corners = (charuco_length - 1) * (charuco_width - 1)


def detect_charuco_board(image):
    """
    Charuco base pose estimation.
    more-or-less copied from - https://mecaruco2.readthedocs.io/en/latest/notebooks_rst/Aruco/sandbox/ludovic/aruco_calibration_rotation.html
    """
    charuco_corners = []
    charuco_ids = []

    # SUB PIXEL CORNER DETECTION CRITERION
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 100, 0.00001)

    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    aruco_square_corners, aruco_square_ids, rejectedImgPoints = cv2.aruco.detectMarkers(
        gray, aruco_dict
    )

    if len(aruco_square_corners) > 0:
        # SUB PIXEL DETECTION
        for this_corner in aruco_square_corners:
            cv2.cornerSubPix(
                gray, this_corner, winSize=(3, 3), zeroZone=(-1, -1), criteria=criteria
            )
        res2 = cv2.aruco.interpolateCornersCharuco(
            aruco_square_corners, aruco_square_ids, gray, board
        )

        if res2[1] is not None and res2[2] is not None and len(res2[1]) > 3:
            charuco_corners = res2[1]
            charuco_ids = res2[2]

    return charuco_corners, charuco_ids, aruco_square_corners, aruco_square_ids


if __name__ == "__main__":
    pathos_mp_helper.freeze_support()

    console = Console()  # create rich console to catch and print exceptions

    this_computer_name = socket.gethostname()

    freemocap_data_path = None
    in_rotation_codes_list = None
    in_cams_to_use_list = None

    if this_computer_name == "jon-hallway-XPS-8930":
        freemocap_data_path = Path("/home/jon/Dropbox/FreeMoCapProject/FreeMocap_Data")
        in_rotation_codes_list = [
            "cv2.ROTATE_90_COUNTERCLOCKWISE",
            "cv2.ROTATE_90_COUNTERCLOCKWISE",
            "cv2.ROTATE_90_CLOCKWISE",
            "cv2.ROTATE_90_CLOCKWISE",
            "cv2.ROTATE_90_CLOCKWISE",
        ]
    elif this_computer_name == "DESKTOP-DCG6K4F":
        freemocap_data_path = Path(
            r"C:\Users\jonma\Dropbox\FreeMoCapProject\FreeMocap_Data"
        )
    elif this_computer_name == "DESKTOP-V3D343U":
        freemocap_data_path = Path(
            r"C:\Users\WindowsPC_Hallway\Dropbox\FreeMoCapProject\FreeMocap_Data"
        )
        in_cams_to_use_list = [0, 1, 2, 3]
    elif this_computer_name == "Jons-MacBook-Pro.local":
        freemocap_data_path = Path("/Users/jon/Dropbox/FreeMoCapProject")

    try:
        multi_cam = FMC_MultiCamera(
            cams_to_use_list=in_cams_to_use_list,
            show_multi_cam_stream_bool=False,
            freemocap_data_folder=str(freemocap_data_path),
            rotation_codes_list=in_rotation_codes_list,
        )
        multi_cam.start()

        console.rule("Launching Multi Cam Viewer")
        current_loop_count = 0
        aruco_square_corners = [[] for _ in range(multi_cam.num_cams)]
        aruco_square_ids = [np.array([]) for _ in range(multi_cam.num_cams)]
        charuco_corners = [np.array([]) for _ in range(multi_cam.num_cams)]
        charuco_ids = [np.array([]) for _ in range(multi_cam.num_cams)]

        camera_pair_list = list(combinations(np.arange(multi_cam.num_cams), 2))
        camera_pair_joint_view_count = [0 for _ in range(len(camera_pair_list))]
        each_cameras_shared_board_view_count_array = np.zeros(
            (multi_cam.num_cams, multi_cam.num_cams)
        )
        each_cameras_shared_board_view_count_total = np.zeros(multi_cam.num_cams)

        charuco_points_from_previous_frames = [[] for _ in range(multi_cam.num_cams)]

        multi_cam_queue_size_list = []

        # #set up plots
        # plt.ion()
        # fig = plt.figure()
        # ax_queue_size  = plt.subplot(121, title='Multi Camera - Queueue Size', xlabel='Frame# I guess/? Maybe Loop#/?', ylabel='Multi frames in queue')
        # ax_shared_views  = plt.subplot(122, title='Shared Charuco Board Views', xlabel='Frame#/?', ylabel='Number of shared views')

        # artist_for_queue_trace = ax_queue_size.plot(0)

        # artists_for_shared_view_traces_list = []
        # for this_cam_num in range(multi_cam.num_cams):
        #     artists_for_shared_view_traces_list.append(ax_shared_views.plot(0, label=f'Cam {this_cam_num}'))
        # ax_shared_views.legend()

        while not multi_cam.exit_event.is_set():
            if not multi_cam.multi_cam_tuple_queue.empty():

                this_multi_cam_tuple = multi_cam.multi_cam_tuple_queue.get()
                current_loop_count += 1

                # multi_cam.save_synchronized_videos(this_multi_cam_tuple)

                # this_multi_cam_image_raw = multi_cam.stitch_multicam_image(this_multi_cam_tuple)

                # if multi_cam._save_multi_cam_to_mp4:
                #     if multi_cam._output_multi_cam_video_object is None:
                #         multi_cam.initialize_multi_cam_output_video(this_multi_cam_image_raw)

                # multi_cam._output_multi_cam_video_object.write(this_multi_cam_image_raw)

                if multi_cam._save_each_cam_to_mp4:
                    if multi_cam._each_cam_video_writer_object_list is None:
                        multi_cam.initialize_each_cam_output_video(this_multi_cam_tuple)

                    for this_cam_num in range(multi_cam.num_cams):
                        multi_cam._each_cam_video_writer_object_list[
                            this_cam_num
                        ].write(this_multi_cam_tuple[this_cam_num][1])

                new_cam_image_list = []
                full_charuco_detected_on_this_frame = [
                    False for _ in range(multi_cam.num_cams)
                ]

                for this_cam_num in range(multi_cam.num_cams):
                    this_cam_tuple = this_multi_cam_tuple[this_cam_num]
                    this_cam_image = this_cam_tuple[1]

                    if current_loop_count % 10 == 0:
                        (
                            charuco_corners[this_cam_num],
                            charuco_ids[this_cam_num],
                            aruco_square_corners[this_cam_num],
                            aruco_square_ids[this_cam_num],
                        ) = detect_charuco_board(this_cam_image)

                    if len(charuco_ids[this_cam_num]) == num_charuco_corners:
                        full_charuco_detected_on_this_frame[this_cam_num] = True

                    # image_w_markers = cv2.aruco.drawDetectedMarkers(this_cam_image, aruco_square_corners[this_cam_num], aruco_square_ids[this_cam_num])

                    image_w_markers = cv2.aruco.drawDetectedCornersCharuco(
                        this_cam_image,
                        np.array(charuco_corners[this_cam_num]),
                        np.array(charuco_ids[this_cam_num]),
                    )

                    this_cam_name = "Camera " + str(this_cam_num)

                    text_to_write_on_this_camera = ""
                    current_cam_corner_count_str = (
                        this_cam_name
                        + ": "
                        + str(len(charuco_ids[this_cam_num]))
                        + " of "
                        + str(num_charuco_corners)
                        + " detected | Full Board: "
                        + str(full_charuco_detected_on_this_frame[this_cam_num])
                    )
                    this_cam_shared_views_str = " | Shared Views: " + str(
                        each_cameras_shared_board_view_count_total[this_cam_num]
                    )
                    text_to_write_on_this_camera = (
                        current_cam_corner_count_str + this_cam_shared_views_str
                    )

                    if this_cam_num == 0:
                        queue_size_str = "Multi-Camera Queue Size: " + str(
                            multi_cam.multi_cam_tuple_queue.qsize()
                        )
                        queue_text_position = (10, 20)
                        cv2.putText(
                            image_w_markers,  # numpy array on which text is written
                            queue_size_str,  # text
                            queue_text_position,  # position at which writing has to start
                            cv2.FONT_HERSHEY_SIMPLEX,  # font family
                            0.5,  # font size
                            (255, 10, 100, 255),  # font color
                            2,
                        )  # font stroke
                        cv2.putText(
                            image_w_markers,  # numpy array on which text is written
                            queue_size_str,  # text
                            queue_text_position,  # position at which writing has to start
                            cv2.FONT_HERSHEY_SIMPLEX,  # font family
                            0.5,  # font size
                            (30, 10, 0, 255),  # font color
                            3,
                        )  # font stroke

                    position = (10, 50)
                    cv2.putText(
                        image_w_markers,  # numpy array on which text is written
                        text_to_write_on_this_camera,  # text
                        position,  # position at which writing has to start
                        cv2.FONT_HERSHEY_SIMPLEX,  # font family
                        0.5,  # font size
                        (30, 10, 0, 255),  # font color
                        3,
                    )  # font stroke
                    cv2.putText(
                        image_w_markers,  # numpy array on which text is written
                        text_to_write_on_this_camera,  # text
                        position,  # position at which writing has to start
                        cv2.FONT_HERSHEY_SIMPLEX,  # font family
                        0.5,  # font size
                        (209, 180, 0, 255),  # font color
                        1,
                    )  # font stroke

                    if full_charuco_detected_on_this_frame[this_cam_num]:
                        cv2.polylines(
                            image_w_markers,
                            np.int32([charuco_corners[this_cam_num]]),
                            True,
                            (0, 100, 255),
                            2,
                        )
                        # for these_corners in charuco_points_from_previous_frames[this_cam_num]:
                        # if len(these_corners)>0:
                        #     cv2.polylines(image_w_markers, np.int32([these_corners[this_cam_num]]), True, (0,100,255,255/2), 2)

                    charuco_points_from_previous_frames[this_cam_num].append(
                        charuco_corners[this_cam_num]
                    )

                    new_cam_image_list.append(this_cam_image)

                # determine paired board views
                for this_cam_num in range(multi_cam.num_cams):
                    if full_charuco_detected_on_this_frame[this_cam_num]:
                        for this_other_camera_num in range(multi_cam.num_cams):
                            if this_other_camera_num != this_cam_num:
                                if full_charuco_detected_on_this_frame[
                                    this_other_camera_num
                                ]:
                                    each_cameras_shared_board_view_count_array[
                                        this_cam_num, this_other_camera_num
                                    ] += 1
                each_cameras_shared_board_view_count_total = np.sum(
                    each_cameras_shared_board_view_count_array, axis=1
                )
                # print(each_cameras_shared_board_view_count)

                for this_pair_num, this_cam_pair in enumerate(camera_pair_list):
                    camera_one_id = this_cam_pair[0]
                    camera_two_id = this_cam_pair[1]

                    if (
                        full_charuco_detected_on_this_frame[camera_one_id]
                        and full_charuco_detected_on_this_frame[camera_two_id]
                    ):
                        camera_pair_joint_view_count[this_pair_num] += 1

                # update plots
                multi_cam_queue_size_list.append(
                    multi_cam.multi_cam_tuple_queue.qsize()
                )

                # artist_for_queue_trace[0].set_data((np.arange(current_loop_count),multi_cam_queue_size_list))
                # plt.pause(0.01)
                # plt.draw()
                # for camNum in range(multi_cam.num_cams):
                #     ax_shared_views.plot(0, label=f'Cam {camNum}')

                multi_cam_image = np.hstack(new_cam_image_list)

                cv2.imshow(multi_cam._rec_name, multi_cam_image)

                key = cv2.waitKey(1)

                if key == 27:  # exit on ESC
                    break

                if (
                    cv2.getWindowProperty(multi_cam._rec_name, cv2.WND_PROP_VISIBLE) < 1
                ):  # break loop if window closed
                    break

        cv2.destroyAllWindows()

        if multi_cam._save_multi_cam_to_mp4:
            multi_cam._output_multi_cam_video_object.release()

        if multi_cam._save_each_cam_to_mp4:
            for this_cam_num in range(multi_cam.num_cams):
                multi_cam._each_cam_video_writer_object_list[this_cam_num].release()

        console.rule("Shutting down MultiCamera Viewer")
        multi_cam.exit_event.set()  # send the 'Exit' signal to everyone.

    except Exception:
        console.print_exception()



import logging
import platform
import time
import traceback

import cv2

from src.cameras.capture.dataclasses.frame_payload import FramePayload
from src.cameras.persistence.video_writer.video_recorder import VideoRecorder
from src.config.webcam_config import WebcamConfig
from src.cameras.capture.opencv_camera.camera_stream_thread_handler import VideoCaptureThread

logger = logging.getLogger(__name__)


class OpenCVCamera:
    """
    Performant implementation of video capture against webcams
    """
    def __init__(self, config: WebcamConfig, session_id: str = None):
        self._config = config
        self._name = f"Camera_{self._config.webcam_id}"
        self._opencv_video_capture_object: cv2.VideoCapture = None
        self._video_recorder: VideoRecorder = None
        self._running_thread: VideoCaptureThread = None
        self._new_frame_ready = False

    @property
    def new_frame_ready(self):
        """
        can be called to determine if the frame returned from  `self.latest_frame` is new or if it has been called before
        """
        return self._new_frame_ready

    @property
    def video_recorder(self):
        return self._video_recorder

    @property
    def name(self):
        return self._name

    @property
    def webcam_id_as_str(self):
        return str(self._config.webcam_id)

    @property
    def current_fps(self):
        return self._running_thread.average_fps

    @property
    def is_capturing_frames(self):
        if not self._running_thread:
            logger.info("Frame Capture thread not running yet")
            return False
        return self._running_thread.is_capturing_frames

    @property
    def current_fps_short(self) -> str:
        return "{:.2f}".format(self._running_thread.average_fps)

    @property
    def latest_frame(self):
        self._new_frame_ready = False
        return self._running_thread.latest_frame

    def connect(self):
        if platform.system() == "Windows":
            cap_backend = cv2.CAP_DSHOW
        else:
            cap_backend = cv2.CAP_ANY

        self._opencv_video_capture_object = cv2.VideoCapture(
            self._config.webcam_id, cap_backend
        )
        self._apply_configuration()
        success, image = self._opencv_video_capture_object.read()

        if not success:
            logger.error(
                "Could not connect to a camera at port# {}".format(
                    self._config.webcam_id
                )
            )
            return success

        try:
            image_height = image.shape[0]
            image_width = image.shape[1]
            self._video_recorder = VideoRecorder(self._name,
                                                 image_width=image_width,
                                                 image_height=image_height)
        except:
            logger.error(f"could not create video recorder for image with shape (image_width={image_width}, image_height={image_height}) ")
            return False

        logger.debug(f"Camera found at port number {self._config.webcam_id}")
        fps_input_stream = int(self._opencv_video_capture_object.get(5))
        logger.debug("FPS of webcam hardware/input stream: {}".format(fps_input_stream))

        return success

    def start_frame_capture_thread(self):
        if self.is_capturing_frames:
            logger.debug(
                f"Already capturing frames for webcam_id: {self.webcam_id_as_str}"
            )
            return
        logger.info(
            f"Beginning frame capture thread for webcam: {self.webcam_id_as_str}"
        )
        self._running_thread = self._create_thread()
        self._running_thread.start()

    def _create_thread(self):
        return VideoCaptureThread(
            get_next_frame=self.get_next_frame,
        )

    @property
    def image_width(self):
        try:
            return int(self._opencv_video_capture_object.get(3))
        except Exception as e:
            raise e


    @property
    def image_height(self):
        try:
            return int(self._opencv_video_capture_object.get(4))
        except Exception as e:
            raise e

    def _apply_configuration(self):
        # set camera stream parameters
        self._opencv_video_capture_object.set(
            cv2.CAP_PROP_EXPOSURE, self._config.exposure
        )
        self._opencv_video_capture_object.set(
            cv2.CAP_PROP_FRAME_WIDTH, self._config.resolution_width
        )
        self._opencv_video_capture_object.set(
            cv2.CAP_PROP_FRAME_HEIGHT, self._config.resolution_height
        )

        self._opencv_video_capture_object.set(
            cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*self._config.fourcc)
        )

    def get_next_frame(self):

        # Why `grab()` not `read()`? see ->
        # https://stackoverflow.com/questions/57716962/difference-between-video-capture-read-and
        # -grab

        if not self._opencv_video_capture_object.grab():
            return FramePayload(False, None, None)

        timestamp_ns = time.time_ns()
        success, image = self._opencv_video_capture_object.retrieve()

        self._new_frame_ready = success
        return FramePayload(success, image, timestamp_ns)

    def stop_frame_capture(self):
        self.close()

    def close(self):
        try:
            self._running_thread.stop()
            while self._running_thread.is_alive():
                # wait for thread to die.
                # TODO: use threading.Event for synchronize mainthread vs other threads
                time.sleep(0.1)
        except:
            logger.error("Printing traceback")
            traceback.print_exc()
        finally:
            logger.info("Closed {}".format(self._name))
            self._opencv_video_capture_object.release()



import bpy
import addon_utils
import numpy as np
from pathlib import Path
import sys
import traceback

#######################################################################
##% Activate necessary addons
addon_utils.enable("io_import_images_as_planes")
addon_utils.enable("rigify")

#######################################################################
##% Session path
# #get session path as command line argument 
argv = sys.argv
argv = argv[argv.index("--") + 1:] 
session_path = argv[0]
if len(argv)>1:
    good_clean_frame_number = int(argv[1])
else:
    good_clean_frame_number = 0
    
print(str(session_path))
session_path = Path(session_path)
path_to_mediapipe_npy = session_path / 'DataArrays' / 'mediaPipeSkel_3d_smoothed.npy'
path_to_mediapipe_3d_reproj = session_path / 'DataArrays' / 'mediaPipeSkel_reprojErr.npy'


#######################################################################
#%% load mediapipe data
#%%
print('loading data')
# %%
mediapipe_skel_fr_mar_dim = np.load(path_to_mediapipe_npy)
mediapipe_skel_fr_mar_dim = mediapipe_skel_fr_mar_dim/1000 #convert to meters

mediapipe_reproj_fr_mar = np.load(path_to_mediapipe_3d_reproj)
mediapipe_reproj_fr_mar = mediapipe_reproj_fr_mar/1000 #convert to meters

body_marker_range = np.arange(0,33)
right_hand_marker_range = np.arange(33, 33+21)
left_hand_marker_range = np.arange(33+21, 33+21+21)
face_marker_range = np.arange(left_hand_marker_range[-1], mediapipe_skel_fr_mar_dim.shape[1])

body_skel_fr_mar_dim = mediapipe_skel_fr_mar_dim[:, body_marker_range, :,]
right_hand_fr_mar_dim = mediapipe_skel_fr_mar_dim[:, right_hand_marker_range, :,]
left_hand_fr_mar_dim = mediapipe_skel_fr_mar_dim[:, left_hand_marker_range, :,]
face_fr_mar_dim = mediapipe_skel_fr_mar_dim[:, face_marker_range, :,]


#%% Session Specific stuff
number_of_frames = mediapipe_skel_fr_mar_dim.shape[0]
start_frame = 1
end_frame = number_of_frames

bpy.context.scene.frame_start = start_frame
bpy.context.scene.frame_end = end_frame


########################
## Find Good Clean Frame with some fiddly nonsense
if good_clean_frame_number == 0:
    frame_nan_counts = []
    frame_mean_reproj_error = []

    for this_frame in range(mediapipe_skel_fr_mar_dim.shape[0]):
        frame_nan_counts.append(np.sum(np.isnan(mediapipe_skel_fr_mar_dim[this_frame,:,0])))
        frame_mean_reproj_error.append(np.nanmean(mediapipe_reproj_fr_mar[this_frame,:]))

    nan_times_vis = np.array(frame_nan_counts)*np.array(frame_mean_reproj_error)
    num_frames  = len(frame_nan_counts)
    # nan_times_vis[0:int(num_frames/5)] = np.nanmax(nan_times_vis)
    # nan_times_vis[-int(num_frames/5):-1] = np.nanmax(nan_times_vis)

    good_clean_frame_number = np.nanargmin(nan_times_vis) # the frame with the fewest nans (i.e. hopefully a frame where all tracked points are visible)

print(f"----good_clean_frame_number: {good_clean_frame_number}----")

#######################################################################
#%% Mediapipe tracked point names
mediapipe_tracked_point_names =  [
    "nose",
    "left_eye_inner",
    "left_eye",
    "left_eye_outer",
    "right_eye_inner",
    "right_eye",
    "right_eye_outer",
    "left_ear",
    "right_ear",
    "mouth_left",
    "mouth_right",
    "left_shoulder",
    "right_shoulder",
    "left_elbow",
    "right_elbow",
    "left_wrist",
    "right_wrist",
    "left_pinky",
    "right_pinky",
    "left_index",
    "right_index",
    "left_thumb",
    "right_thumb",
    "left_hip",
    "right_hip",
    "left_knee",
    "right_knee",
    "left_ankle",
    "right_ankle",
    "left_heel",
    "right_heel",
    "left_foot_index",
    "right_foot_index",
    "right_hand_wrist",
    "right_hand_thumb_cmc",
    "right_hand_thumb_mcp",
    "right_hand_thumb_ip",
    "right_hand_thumb_tip",
    "right_hand_index_finger_mcp",
    "right_hand_index_finger_pip",
    "right_hand_index_finger_dip",
    "right_hand_index_finger_tip",
    "right_hand_middle_finger_mcp",
    "right_hand_middle_finger_pip",
    "right_hand_middle_finger_dip",
    "right_hand_middle_finger_tip",
    "right_hand_ring_finger_mcp",
    "right_hand_ring_finger_pip",
    "right_hand_ring_finger_dip",
    "right_hand_ring_finger_tip",
    "right_hand_pinky_finger_mcp",
    "right_hand_pinky_finger_pip",
    "right_hand_pinky_finger_dip",
    "right_hand_pinky_finger_tip",
    "left_hand_wrist",
    "left_hand_thumb_cmc",
    "left_hand_thumb_mcp",
    "left_hand_thumb_ip",
    "left_hand_thumb_tip",
    "left_hand_index_finger_mcp",
    "left_hand_index_finger_pip",
    "left_hand_index_finger_dip",
    "left_hand_index_finger_tip",
    "left_hand_middle_finger_mcp",
    "left_hand_middle_finger_pip",
    "left_hand_middle_finger_dip",
    "left_hand_middle_finger_tip",
    "left_hand_ring_finger_mcp",
    "left_hand_ring_finger_pip",
    "left_hand_ring_finger_dip",
    "left_hand_ring_finger_tip",
    "left_hand_pinky_finger_mcp",
    "left_hand_pinky_finger_pip",
    "left_hand_pinky_finger_dip",
    "left_hand_pinky_finger_tip",
    ]


#####################################################################
## Blender Stuff
#####################################################################

try:
    #####################################################################
    ###%% clear scene
    print('clearing scene')
    try:
        bpy.ops.object.mode_set(mode='OBJECT')
    except:
        pass

    try:
        bpy.ops.object.select_all(action='SELECT') # select all objects
        bpy.ops.object.delete(use_global=True) # delete all objects from all scenes
    except:
        pass

    #######################################################################
    #%% create world origin

    bpy.ops.object.empty_add(type='ARROWS')
    world_origin_axes = bpy.context.active_object
    world_origin_axes.name = 'world_origin' #will stay at origin
    # inspect(world_origin_axes, methods=True)

    bpy.ops.object.empty_add(type='ARROWS')
    freemocap_origin_axes = bpy.context.active_object
    freemocap_origin_axes.name = 'freemocap_origin_axes' #will translate to put skelly on ground symmetric-ish about origin


    #######################################################################
    #%% load empties
    print('loading {} empties on {} frames'.format(len(mediapipe_tracked_point_names), number_of_frames))

    empty_size = .01

    for this_point_index, this_point_name in enumerate(mediapipe_tracked_point_names):
        print(f'loading {this_point_name}...')
        bpy.ops.object.empty_add(type='SPHERE')
        this_empty = bpy.context.active_object
        this_empty.name = this_point_name

        if 'face' in this_point_name:
            this_empty.scale = [empty_size/4]*3
        if 'hand' in this_point_name:
            this_empty.scale = [empty_size/2]*3
        else:
            this_empty.scale = [empty_size]*3

        this_empty.parent = freemocap_origin_axes

        this_point_fr_xyz = mediapipe_skel_fr_mar_dim[:,this_point_index,:]

        for frame_num in range(end_frame):

            this_empty.location = [this_point_fr_xyz[frame_num,0],
                                    this_point_fr_xyz[frame_num,1],
                                    this_point_fr_xyz[frame_num,2]]
            bpy.context.view_layer.update()

            this_empty.keyframe_insert(data_path='location',frame=frame_num)

    #######################################################################
    #%% create virtual markers
    left_ear_index = 7
    right_ear_index = 8
    #######################################################################
    print('creating virtual markers')
    print('head_center - midway between left and right ears')
    hips_xyz =  (mediapipe_skel_fr_mar_dim[:,left_ear_index,:] + mediapipe_skel_fr_mar_dim[:,right_ear_index,:])/2
    bpy.ops.object.empty_add(type='SPHERE')
    this_empty = bpy.context.active_object
    this_empty.name = 'head_center'
    this_empty.scale = [empty_size]*3
    this_empty.parent = freemocap_origin_axes
    mediapipe_tracked_point_names.append(this_empty.name)

    for frame_num in range(end_frame):
        this_empty.location = [hips_xyz[frame_num,0],
                            hips_xyz[frame_num,1],
                            hips_xyz[frame_num,2]]
        this_empty.keyframe_insert(data_path='location',frame=frame_num)

    print('neck_center - midway between left and right shoulders')
    left_shoulder_index = 11
    right_shoulder_index = 12
    neck_xyz =  (mediapipe_skel_fr_mar_dim[:,left_shoulder_index,:] + mediapipe_skel_fr_mar_dim[:,right_shoulder_index,:])/2
    bpy.ops.object.empty_add(type='PLAIN_AXES')
    this_empty = bpy.context.active_object
    this_empty.name = 'neck_center'
    this_empty.scale = [empty_size]*3
    this_empty.parent = freemocap_origin_axes
    mediapipe_tracked_point_names.append(this_empty.name)

    for frame_num in range(end_frame):
        this_empty.location = [neck_xyz[frame_num,0],
                            neck_xyz[frame_num,1],
                            neck_xyz[frame_num,2]]
        this_empty.keyframe_insert(data_path='location',frame=frame_num)



    print('hip_center - midway between left and right hips')
    left_hip_index = 23
    right_hip_index = 24
    hips_xyz =  (mediapipe_skel_fr_mar_dim[:,left_hip_index,:] + mediapipe_skel_fr_mar_dim[:,right_hip_index,:])/2
    bpy.ops.object.empty_add(type='PLAIN_AXES')
    this_empty = bpy.context.active_object
    this_empty.name = 'hip_center'
    this_empty.scale = [empty_size]*3
    this_empty.parent = freemocap_origin_axes
    mediapipe_tracked_point_names.append(this_empty.name)

    for frame_num in range(end_frame):
        this_empty.location = [hips_xyz[frame_num,0],
                            hips_xyz[frame_num,1],
                            hips_xyz[frame_num,2]]
        this_empty.keyframe_insert(data_path='location',frame=frame_num)


    print('chest_center - mean of R/L shoulders and R/L hips')
    chest_xyz =  (neck_xyz + hips_xyz)/2
    bpy.ops.object.empty_add(type='PLAIN_AXES')
    this_empty = bpy.context.active_object
    this_empty.name = 'chest_center'
    this_empty.scale = [empty_size]*3
    this_empty.parent = freemocap_origin_axes
    mediapipe_tracked_point_names.append(this_empty.name)

    for frame_num in range(end_frame):
        this_empty.location = [chest_xyz[frame_num,0],
                            chest_xyz[frame_num,1],
                            chest_xyz[frame_num,2]]
        this_empty.keyframe_insert(data_path='location',frame=frame_num)

    print('Done loading empties :D')


    ######################################################################
    ## Fit Rigify metarig armatures to empties

    bpy.context.scene.frame_set(good_clean_frame_number)

    #each bone  name has to match one in the Blender Rigify Human Metarig, or it will be ignored
    metarig_bone_empty_pairs_dict_of_dicts = {
    "hand.L": {
        'head': 'left_wrist',
        'tail': 'left_index',
    },
    "forearm.L": {
        'head': 'left_elbow',
        'tail': 'left_wrist',
    },
    "upper_arm.L": {
        'head': 'left_shoulder',
        'tail': 'left_elbow',
    },
    "shoulder.L": {
        'head': 'neck_center',
        'tail': 'left_shoulder',
    },
    "hand.R": {
        'head': 'right_wrist',
        'tail': 'right_index',
    },
    "forearm.R": {
        'head': 'right_elbow',
        'tail': 'right_wrist',
    },
    "upper_arm.R": {
        'head': 'right_shoulder',
        'tail': 'right_elbow',
    },
    "shoulder.R": {
        'head': 'neck_center',
        'tail': 'right_shoulder',
    },
    "thigh.L": {
        'head': 'left_hip',
        'tail': 'left_knee',
    },
    "shin.L": {
        'head': 'left_knee',
        'tail': 'left_ankle',
    },
    "heel.02.L": {
        'head': 'left_heel',
        'tail': 'left_ankle',
    },
    "foot.L": {
        'head': 'left_ankle',
        'tail': 'left_foot_index',
    },
    "toe.L": {
        'head': 'left_foot_index',
        'tail': 'left_heel',
    },
    "thigh.R": {
        'head': 'right_hip',
        'tail': 'right_knee',
    },
    "shin.R": {
        'head': 'right_knee',
        'tail': 'right_ankle',
    },
    "heel.02.R": {
        'head': 'right_heel',
        'tail': 'right_ankle',
    },
    "foot.R": {
        'head': 'right_ankle',
        'tail': 'right_foot_index',
    },
    "toe.R": {
        'head': 'right_foot_index',
        'tail': 'right_heel',
    },
    "pelvis.L": {
        'head': 'hip_center',
        'tail': 'left_hip',
    },
    "pelvis.R": {
        'head': 'hip_center',
        'tail': 'right_hip',
    },
    "spine": {
        'head': 'hip_center',
        'tail': 'chest_center',
    },
    "spine.001": {
        'head': 'chest_center',
        'tail': 'neck_center',
    },
    "spine.002": {
        'head': 'neck_center',
        'tail': 'head_center',
    },
    "spine.003": {
        'head': 'head_center',
        'tail': 'nose',
    },
    "palm.01.R": {
        'head': 'right_wrist',
        'tail': 'right_hand_index_finger_mcp',
    },
    "palm.02.R": {
        'head': 'right_wrist',
        'tail': 'right_hand_middle_finger_mcp',
    },
    "palm.03.R": {
        'head': 'right_wrist',
        'tail': 'right_hand_ring_finger_mcp',
    },
    "palm.04.R": {
        'head': 'right_wrist',
        'tail': 'right_hand_pinky_finger_mcp',
    },

    "thumb.01.R": {
        'head': 'right_wrist',
        'tail': 'right_hand_thumb_mcp',
    },
    "thumb.02.R": {
        'head': 'right_hand_thumb_mcp',
        'tail': 'right_hand_thumb_ip',
    },
    "thumb.03.R": {
        'head': 'right_hand_thumb_ip',
        'tail': 'right_hand_thumb_tip',
    },

    "f_index.01.R": {
        'head': 'right_hand_index_finger_mcp',
        'tail': 'right_hand_index_finger_pip',
    },
    "f_index.02.R": {
        'head': 'right_hand_index_finger_pip',
        'tail': 'right_hand_index_finger_dip',
    },
    "f_index.03.R": {
        'head': 'right_hand_index_finger_dip',
        'tail': 'right_hand_index_finger_tip',
    },

    "f_middle.01.R": {
        'head': 'right_hand_middle_finger_mcp',
        'tail': 'right_hand_middle_finger_pip',
    },
    "f_middle.02.R": {
        'head': 'right_hand_middle_finger_pip',
        'tail': 'right_hand_middle_finger_dip',
    },
    "f_middle.03.R": {
        'head': 'right_hand_middle_finger_dip',
        'tail': 'right_hand_middle_finger_tip',
    },

    "f_ring.01.R": {
        'head': 'right_hand_ring_finger_mcp',
        'tail': 'right_hand_ring_finger_pip',
    },
    "f_ring.02.R": {
        'head': 'right_hand_ring_finger_pip',
        'tail': 'right_hand_ring_finger_dip',
    },
    "f_ring.03.R": {
        'head': 'right_hand_ring_finger_dip',
        'tail': 'right_hand_ring_finger_tip',
    },

    "f_pinky.01.R": {
        'head': 'right_hand_pinky_finger_mcp',
        'tail': 'right_hand_pinky_finger_pip',
    },
    "f_pinky.02.R": {
        'head': 'right_hand_pinky_finger_pip',
        'tail': 'right_hand_pinky_finger_dip',
    },
    "f_pinky.03.R": {
        'head': 'right_hand_pinky_finger_dip',
        'tail': 'right_hand_pinky_finger_tip',
    },
    "thumb.01.L": {
        'head': 'left_wrist',
        'tail': 'left_hand_thumb_mcp',
    },
    "thumb.02.L": {
        'head': 'left_hand_thumb_mcp',
        'tail': 'left_hand_thumb_ip',
    },
    "thumb.03.L": {
        'head': 'left_hand_thumb_ip',
        'tail': 'left_hand_thumb_tip',
    },
    "palm.01.L": {
        'head': 'left_wrist',
        'tail': 'left_hand_index_finger_mcp',
    },
    "palm.02.L": {
        'head': 'left_wrist',
        'tail': 'left_hand_middle_finger_mcp',
    },
    "palm.03.L": {
        'head': 'left_wrist',
        'tail': 'left_hand_ring_finger_mcp',
    },
    "palm.04.L": {
        'head': 'left_wrist',
        'tail': 'left_hand_pinky_finger_mcp',
    },
    "f_index.01.L": {
        'head': 'left_hand_index_finger_mcp',
        'tail': 'left_hand_index_finger_pip',
    },
    "f_index.02.L": {
        'head': 'left_hand_index_finger_pip',
        'tail': 'left_hand_index_finger_dip',
    },
    "f_index.03.L": {
        'head': 'left_hand_index_finger_dip',
        'tail': 'left_hand_index_finger_tip',
    },

    "f_middle.01.L": {
        'head': 'left_hand_middle_finger_mcp',
        'tail': 'left_hand_middle_finger_pip',
    },
    "f_middle.02.L": {
        'head': 'left_hand_middle_finger_pip',
        'tail': 'left_hand_middle_finger_dip',
    },
    "f_middle.03.L": {
        'head': 'left_hand_middle_finger_dip',
        'tail': 'left_hand_middle_finger_tip',
    },
    "f_ring.01.L": {
        'head': 'left_hand_ring_finger_mcp',
        'tail': 'left_hand_ring_finger_pip',
    },
    "f_ring.02.L": {
        'head': 'left_hand_ring_finger_pip',
        'tail': 'left_hand_ring_finger_dip',
    },
    "f_ring.03.L": {
        'head': 'left_hand_ring_finger_dip',
        'tail': 'left_hand_ring_finger_tip',
    },
    "f_pinky.01.L": {
        'head': 'left_hand_pinky_finger_mcp',
        'tail': 'left_hand_pinky_finger_pip',
    },
    "f_pinky.02.L": {
        'head': 'left_hand_pinky_finger_pip',
        'tail': 'left_hand_pinky_finger_dip',
    },
    "f_pinky.03.L": {
        'head': 'left_hand_pinky_finger_dip',
        'tail': 'left_hand_pinky_finger_tip',
    },

    }


    #%% create metarig
    bpy.ops.object.armature_human_metarig_add()
    this_metarig = bpy.context.active_object

    #%% loop through dictionary - set head and tail of each bone's to the relevant empty
    # bpy.ops.object.posemode_toggle()
    try:
        bpy.ops.object.mode_set(mode='EDIT')
    except:
        pass

    for this_bone in list(this_metarig.data.bones):
        this_bone_name = this_bone.name
        print(f'--{this_bone_name}--')

        this_edit_bone= this_metarig.data.edit_bones[this_bone_name] # note this is using EDIT_bones, not regular bones b/c blender is a wild wacky place

        if this_bone_name in metarig_bone_empty_pairs_dict_of_dicts:
            this_bone_dict = metarig_bone_empty_pairs_dict_of_dicts[this_bone_name]

            this_head_empty_name = this_bone_dict['head']
            this_tail_empty_name = this_bone_dict['tail']
            print('setting (bone:head->tail){}:{}->{})'.format(this_bone_name, this_head_empty_name, this_tail_empty_name))

            if this_tail_empty_name is not None:
                print(f'setting bone: {this_bone_name}: head: {bpy.data.objects[this_head_empty_name].location}, tail: {bpy.data.objects[this_tail_empty_name].location}')
                this_edit_bone.head = bpy.data.objects[this_head_empty_name].location
                this_edit_bone.tail = bpy.data.objects[this_tail_empty_name].location
            else: #remove bones with no tail (and/or head) until we figure out a better plan
                print(f'removing bone: {this_bone_name}')
                this_metarig.data.edit_bones.remove(this_edit_bone)
        else:
            print(f'removing bone: {this_bone_name}')
            this_metarig.data.edit_bones.remove(this_edit_bone)

    try:
        bpy.ops.object.mode_set(mode='OBJECT')
    except:
        pass




    #######################################################
    #### Damped track bones to empties
    bpy.ops.object.mode_set(mode='OBJECT')
    armature_name = 'metarig'
    this_armature = bpy.data.objects[armature_name]

    rig_constraint_dict_of_dicts = {
        "spine": {
            'COPY_LOCATION': 'hip_center',
            'DAMPED_TRACK': 'chest_center',
        },
        "spine.001": {
            'DAMPED_TRACK': 'neck_center',
        },
        "spine.002": {
            'DAMPED_TRACK': 'head_center',
        },
        "spine.003": {
            'DAMPED_TRACK': 'nose',
        },
        "pelvis.L": {
            'DAMPED_TRACK': 'left_hip',
        },
        "thigh.L": {
            'COPY_LOCATION': 'left_hip',
            'DAMPED_TRACK': 'left_knee',
        },
        "shin.L": {
            'DAMPED_TRACK': 'left_ankle',
        },
        "foot.L": {
            'DAMPED_TRACK': 'left_foot_index',
        },
        "heel.02.L": {
            'DAMPED_TRACK': 'left_ankle',
        },
        "shoulder.L": {
            'COPY_LOCATION': 'neck_center',
            'DAMPED_TRACK': 'left_shoulder',
        },
        "upper_arm.L": {
            'DAMPED_TRACK': 'left_elbow',
        },
        "forearm.L": {
            'DAMPED_TRACK': 'left_wrist',
        },
        "hand.L": {
            'DAMPED_TRACK': 'left_index',
        },
        "pelvis.R": {
            'DAMPED_TRACK': 'right_hip',
        },
        "thigh.R": {
            'COPY_LOCATION': 'right_hip',
            'DAMPED_TRACK': 'right_knee',
        },
        "shin.R": {
            'DAMPED_TRACK': 'right_ankle',
        },
        "foot.R": {
            'DAMPED_TRACK': 'right_foot_index',
        },
        "heel.02.R": {
            'DAMPED_TRACK': 'right_ankle',
        },
        "shoulder.R": {
            'COPY_LOCATION': 'neck_center',
            'DAMPED_TRACK': 'right_shoulder',
        },
        "upper_arm.R": {
            'DAMPED_TRACK': 'right_elbow',
        },
        "forearm.R": {
            'DAMPED_TRACK': 'right_wrist',
        },
        "hand.R": {
            'DAMPED_TRACK': 'right_index',
        },
        "thumb.01.R": {
            'COPY_LOCATION': 'right_wrist',
            'DAMPED_TRACK': 'right_hand_thumb_mcp',
        },
        "thumb.02.R": {
            'DAMPED_TRACK': 'right_hand_thumb_ip',
        },
        "thumb.02.R": {
            'DAMPED_TRACK': 'right_hand_thumb_tip',
        },
        "palm.01.R": {
            'COPY_LOCATION': 'right_wrist',
            'DAMPED_TRACK': 'right_hand_index_finger_mcp',
        },
        "f_index.01.R": {
            'DAMPED_TRACK': 'right_hand_index_finger_pip',
        },
        "f_index.02.R": {
            'DAMPED_TRACK': 'right_hand_index_finger_dip',
        },
        "f_index.03.R": {
            'DAMPED_TRACK': 'right_hand_index_finger_tip',
        },
        "palm.02.R": {
            'COPY_LOCATION': 'right_wrist',
            'DAMPED_TRACK': 'right_hand_middle_finger_mcp',
        },
        "f_middle.01.R": {
            'DAMPED_TRACK': 'right_hand_middle_finger_pip',
        },
        "f_middle.02.R": {
            'DAMPED_TRACK': 'right_hand_middle_finger_dip',
        },
        "f_middle.03.R": {
            'DAMPED_TRACK': 'right_hand_middle_finger_tip',
        },
        "palm.03.R": {
            'COPY_LOCATION': 'right_wrist',
            'DAMPED_TRACK': 'right_hand_ring_finger_mcp',
        },
        "f_ring.01.R": {
            'DAMPED_TRACK': 'right_hand_ring_finger_pip',
        },
        "f_ring.02.R": {
            'DAMPED_TRACK': 'right_hand_ring_finger_dip',
        },
        "f_ring.03.R": {
            'DAMPED_TRACK': 'right_hand_ring_finger_tip',
        },
        "palm.04.R": {
            'COPY_LOCATION': 'right_wrist',
            'DAMPED_TRACK': 'right_hand_pinky_finger_mcp',
        },
        "f_pinky.01.R": {
            'DAMPED_TRACK': 'right_hand_pinky_finger_pip',
        },
        "f_pinky.02.R": {
            'DAMPED_TRACK': 'right_hand_pinky_finger_dip',
        },
        "f_pinky.03.R": {
            'DAMPED_TRACK': 'right_hand_pinky_finger_tip',
        },
        "thumb.01.L": {
            'COPY_LOCATION': 'left_wrist',
            'DAMPED_TRACK': 'left_hand_thumb_mcp',
        },
        "thumb.02.L": {
            'DAMPED_TRACK': 'left_hand_thumb_ip',
        },
        "thumb.03.L": {
            'DAMPED_TRACK': 'left_hand_thumb_tip',
        },
        "palm.01.L": {
            'COPY_LOCATION': 'left_wrist',
            'DAMPED_TRACK': 'left_hand_index_finger_mcp',
        },
        "f_index.01.L": {
            'DAMPED_TRACK': 'left_hand_index_finger_pip',
        },
        "f_index.02.L": {
            'DAMPED_TRACK': 'left_hand_index_finger_dip',
        },
        "f_index.03.L": {
            'DAMPED_TRACK': 'left_hand_index_finger_tip',
        },
        "palm.02.L": {
            'COPY_LOCATION': 'left_wrist',
            'DAMPED_TRACK': 'left_hand_middle_finger_mcp',
        },
        "f_middle.01.L": {
            'DAMPED_TRACK': 'left_hand_middle_finger_pip',
        },
        "f_middle.02.L": {
            'DAMPED_TRACK': 'left_hand_middle_finger_dip',
        },
        "f_middle.03.L": {
            'DAMPED_TRACK': 'left_hand_middle_finger_tip',
        },
        "palm.03.L": {
            'COPY_LOCATION': 'left_wrist',
            'DAMPED_TRACK': 'left_hand_ring_finger_mcp',
        },
        "f_ring.01.L": {
            'DAMPED_TRACK': 'left_hand_ring_finger_pip',
        },
        "f_ring.02.L": {
            'DAMPED_TRACK': 'left_hand_ring_finger_dip',
        },
        "f_ring.03.L": {
            'DAMPED_TRACK': 'left_hand_ring_finger_tip',
        },
        "palm.04.L": {
            'COPY_LOCATION': 'left_wrist',
            'DAMPED_TRACK': 'left_hand_pinky_finger_mcp',
        },
        "f_pinky.01.L": {
            'DAMPED_TRACK': 'left_hand_pinky_finger_pip',
        },
        "f_pinky.02.L": {
            'DAMPED_TRACK': 'left_hand_pinky_finger_dip',
        },
        "f_pinky.03.L": {
            'DAMPED_TRACK': 'left_hand_pinky_finger_tip',
        },
    }

    #loop through dictionary applying  constraints
    try:
        bpy.ops.object.mode_set(mode='POSE')
    except:
        pass

    for this_bone_name, this_bone_dict in rig_constraint_dict_of_dicts.items():
        print(f'---Setting constraints for bone:{this_bone_name}---')

        for this_constraint_name, this_constraint_target_empty_name in this_bone_dict.items():
            print(f'constraint: {this_constraint_name} with target:{this_constraint_target_empty_name}')

            this_bone= this_armature.pose.bones[this_bone_name]

            this_constraint = this_bone.constraints.new(type=this_constraint_name)
            this_constraint.name = this_constraint_name
            this_constraint.target = bpy.data.objects[this_constraint_target_empty_name] #point constraint at relevant empty object

    ##############
    ## create extra bones

    extra_bone_empty_pairs_dict_of_dicts = {
    "hand.L": {
        'head': 'left_wrist',
        'tail': 'left_index',
    },
    "forearm.L": {
        'head': 'left_elbow',
        'tail': 'left_wrist',
    },
    }

    ###################################################################
    #### Create Stick Figure Mesh

    stick_figure_mesh_edge_pairs_dict_of_dicts = {

        "forearm.L": {
            'head': 'left_elbow',
            'tail': 'left_wrist',
        },
        "upper_arm.L": {
            'head': 'left_shoulder',
            'tail': 'left_elbow',
        },
        "shoulder.L": {
            'head': 'neck_center',
            'tail': 'left_shoulder',
        },
        "forearm.R": {
            'head': 'right_elbow',
            'tail': 'right_wrist',
        },
        "upper_arm.R": {
            'head': 'right_shoulder',
            'tail': 'right_elbow',
        },
        "shoulder.R": {
            'head': 'neck_center',
            'tail': 'right_shoulder',
        },
        "thigh.L": {
            'head': 'left_hip',
            'tail': 'left_knee',
        },
        "shin.L": {
            'head': 'left_knee',
            'tail': 'left_ankle',
        },
        "heel.02.L": {
            'head': 'left_heel',
            'tail': 'left_ankle',
        },
        "foot.L": {
            'head': 'left_ankle',
            'tail': 'left_foot_index',
        },
        "toe.L": {
            'head': 'left_foot_index',
            'tail': 'left_heel',
        },
        "thigh.R": {
            'head': 'right_hip',
            'tail': 'right_knee',
        },
        "shin.R": {
            'head': 'right_knee',
            'tail': 'right_ankle',
        },
        "heel.02.R": {
            'head': 'right_heel',
            'tail': 'right_ankle',
        },
        "foot.R": {
            'head': 'right_ankle',
            'tail': 'right_foot_index',
        },
        "toe.R": {
            'head': 'right_foot_index',
            'tail': 'right_heel',
        },
        "pelvis.L": {
            'head': 'hip_center',
            'tail': 'left_hip',
        },
        "pelvis.R": {
            'head': 'hip_center',
            'tail': 'right_hip',
        },
        "spine": {
            'head': 'hip_center',
            'tail': 'chest_center',
        },
        "spine.001": {
            'head': 'chest_center',
            'tail': 'neck_center',
        },
        "spine.002": {
            'head': 'neck_center',
            'tail': 'head_center',
        },
        "spine.003": {
            'head': 'head_center',
            'tail': 'nose',
        },
        "palm.01.R": {
            'head': 'right_wrist',
            'tail': 'right_hand_index_finger_mcp',
        },
        "palm.02.R": {
            'head': 'right_wrist',
            'tail': 'right_hand_middle_finger_mcp',
        },
        "palm.03.R": {
            'head': 'right_wrist',
            'tail': 'right_hand_ring_finger_mcp',
        },
        "palm.04.R": {
            'head': 'right_wrist',
            'tail': 'right_hand_pinky_finger_mcp',
        },

        "thumb.01.R": {
            'head': 'right_wrist',
            'tail': 'right_hand_thumb_cmc',
        },
        "thumb.02.R": {
            'head': 'right_hand_thumb_cmc',
            'tail': 'right_hand_thumb_mcp',
        },
        "thumb.03.R": {
            'head': 'right_hand_thumb_mcp',
            'tail': 'right_hand_thumb_ip',
        },
        "thumb.04.R": {
            'head': 'right_hand_thumb_ip',
            'tail': 'right_hand_thumb_tip',
        },
        "f_index.01.R": {
            'head': 'right_hand_index_finger_mcp',
            'tail': 'right_hand_index_finger_pip',
        },
        "f_index.02.R": {
            'head': 'right_hand_index_finger_pip',
            'tail': 'right_hand_index_finger_dip',
        },
        "f_index.03.R": {
            'head': 'right_hand_index_finger_dip',
            'tail': 'right_hand_index_finger_tip',
        },
        "f_middle.01.R": {
            'head': 'right_hand_middle_finger_mcp',
            'tail': 'right_hand_middle_finger_pip',
        },
        "f_middle.02.R": {
            'head': 'right_hand_middle_finger_pip',
            'tail': 'right_hand_middle_finger_dip',
        },
        "f_middle.03.R": {
            'head': 'right_hand_middle_finger_dip',
            'tail': 'right_hand_middle_finger_tip',
        },
        "f_ring.01.R": {
            'head': 'right_hand_ring_finger_mcp',
            'tail': 'right_hand_ring_finger_pip',
        },
        "f_ring.02.R": {
            'head': 'right_hand_ring_finger_pip',
            'tail': 'right_hand_ring_finger_dip',
        },
        "f_ring.03.R": {
            'head': 'right_hand_ring_finger_dip',
            'tail': 'right_hand_ring_finger_tip',
        },

        "f_pinky.01.R": {
            'head': 'right_hand_pinky_finger_mcp',
            'tail': 'right_hand_pinky_finger_pip',
        },
        "f_pinky.02.R": {
            'head': 'right_hand_pinky_finger_pip',
            'tail': 'right_hand_pinky_finger_dip',
        },
        "f_pinky.03.R": {
            'head': 'right_hand_pinky_finger_dip',
            'tail': 'right_hand_pinky_finger_tip',
        },    
        "thumb.01.L": {
            'head': 'left_wrist',
            'tail': 'left_hand_thumb_cmc',
        },
        "thumb.02.L": {
            'head': 'left_hand_thumb_cmc',
            'tail': 'left_hand_thumb_mcp',
        },
        "thumb.03.L": {
            'head': 'left_hand_thumb_mcp',
            'tail': 'left_hand_thumb_ip',
        },
        "thumb.04.L": {
            'head': 'left_hand_thumb_ip',
            'tail': 'left_hand_thumb_tip',
        },
        "palm.01.L": {
            'head': 'left_wrist',
            'tail': 'left_hand_index_finger_mcp',
        },
        "palm.02.L": {
            'head': 'left_wrist',
            'tail': 'left_hand_middle_finger_mcp',
        },
        "palm.03.L": {
            'head': 'left_wrist',
            'tail': 'left_hand_ring_finger_mcp',
        },
        "palm.04.L": {
            'head': 'left_wrist',
            'tail': 'left_hand_pinky_finger_mcp',
        },
        "f_index.01.L": {
            'head': 'left_hand_index_finger_mcp',
            'tail': 'left_hand_index_finger_pip',
        },
        "f_index.02.L": {
            'head': 'left_hand_index_finger_pip',
            'tail': 'left_hand_index_finger_dip',
        },
        "f_index.03.L": {
            'head': 'left_hand_index_finger_dip',
            'tail': 'left_hand_index_finger_tip',
        },
        "f_middle.01.L": {
            'head': 'left_hand_middle_finger_mcp',
            'tail': 'left_hand_middle_finger_pip',
        },
        "f_middle.02.L": {
            'head': 'left_hand_middle_finger_pip',
            'tail': 'left_hand_middle_finger_dip',
        },
        "f_middle.03.L": {
            'head': 'left_hand_middle_finger_dip',
            'tail': 'left_hand_middle_finger_tip',
        },
        "f_ring.01.L": {
            'head': 'left_hand_ring_finger_mcp',
            'tail': 'left_hand_ring_finger_pip',
        },
        "f_ring.02.L": {
            'head': 'left_hand_ring_finger_pip',
            'tail': 'left_hand_ring_finger_dip',
        },
        "f_ring.03.L": {
            'head': 'left_hand_ring_finger_dip',
            'tail': 'left_hand_ring_finger_tip',
        },
        "f_pinky.01.L": {
            'head': 'left_hand_pinky_finger_mcp',
            'tail': 'left_hand_pinky_finger_pip',
        },
        "f_pinky.02.L": {
            'head': 'left_hand_pinky_finger_pip',
            'tail': 'left_hand_pinky_finger_dip',
        },
        "f_pinky.03.L": {
            'head': 'left_hand_pinky_finger_dip',
            'tail': 'left_hand_pinky_finger_tip',
        },
        }

    mediapipe_virtual_markers_weights_dict = {
        'head_center':{
            'right_ear': .5,
            'left_ear': .5
        },
        'neck_center':{
            'right_shoulder': .5,
            'left_shoulder': .5
        },
        'hip_center':{
            'right_hip': .5,
            'left_hip': .5
        },
        'chest_center':{
            'right_shoulder': .25,
            'left_shoulder': .25,
            'right_hip': .25,
            'left_hip': .25,
        },
    }

    print('loading data as stick figure mesh \o/')
    #######################################################################

    #%% create virtual marker points (redundant from above, I think ? )



    current_frame = bpy.data.scenes[0].frame_current
    virtual_point_index_dict = {}
    for this_virtual_marker_dict in mediapipe_virtual_markers_weights_dict.items():
        this_virtual_marker_name = this_virtual_marker_dict[0]
        these_weighted_points = []
        for this_weighted_point_dict in this_virtual_marker_dict[1].items():
            this_point_name = this_weighted_point_dict[0]

            this_weight = this_weighted_point_dict[1]
            these_weighted_points.append(
                mediapipe_skel_fr_mar_dim[
                    :,
                    mediapipe_tracked_point_names.index(this_point_name),
                    :
                    ]*this_weight
                )
        this_virtual_point_fr_xyz = np.sum(these_weighted_points, axis=0)
        this_virtual_point_fr_mar_xyz = np.expand_dims(this_virtual_point_fr_xyz, axis=1)
        mediapipe_skel_fr_mar_dim = np.append(mediapipe_skel_fr_mar_dim,this_virtual_point_fr_mar_xyz, axis=1)
        virtual_point_index_dict[this_virtual_marker_name] = mediapipe_skel_fr_mar_dim.shape[1]-1
        print(f'Virtual marker: {this_virtual_marker_name} created at marker index {mediapipe_skel_fr_mar_dim.shape[1]-1}')


    #%% define edges


    edges = []
    #regular data
    for this_bone_dict in stick_figure_mesh_edge_pairs_dict_of_dicts.items():
        edge_name = this_bone_dict[0]
        head_name = this_bone_dict[1]['head']
        tail_name = this_bone_dict[1]['tail']
        

        try:
            
            head_index = mediapipe_tracked_point_names.index(head_name)
            tail_index = mediapipe_tracked_point_names.index(tail_name)
            
            if 'center' in head_name: #virtual markers are acting weird so we're brute forcing it :-/ this will fail if there is a virtual marker without 'center' in the name.... 
                head_index = virtual_point_index_dict[head_name]
            if 'center' in tail_name:
                tail_index = virtual_point_index_dict[tail_name]

            this_edge =(head_index, tail_index)
            
            edges.append(this_edge)
            print(f'edge created for bone: ({this_bone_dict[0]} : {head_name}-{tail_name}), indicies({this_edge[0]},{this_edge[1]})')
        except:
            print(f'edge FAILED for for bone:{this_bone_dict[0]} : {head_name}-{tail_name}')
            pass
        
    print('edges created!')


    ##############################
    #%% Create mesh
    try:
        bpy.ops.object.mode_set(mode='OBJECT')
    except:
        pass

    print('loading verticies')
    vertices = mediapipe_skel_fr_mar_dim[current_frame,:,:]

    #edges defined above
    faces = []
    stick_figure_mesh = bpy.data.meshes.new('stick_figure_mesh')
    stick_figure_mesh.from_pydata(vertices, edges, faces)
    stick_figure_mesh.update()
    # make object from mesh
    stick_figure_mesh_object = bpy.data.objects.new('stick_figure_mesh', stick_figure_mesh)
    # make collection
    mesh_collection = bpy.data.collections.new('mesh_collection')
    bpy.context.scene.collection.children.link(mesh_collection)
    # add object to scene collection
    mesh_collection.objects.link(stick_figure_mesh_object)

    bpy.context.view_layer.objects.active = bpy.data.objects['stick_figure_mesh']

    try:
        bpy.ops.object.mode_set(mode='EDIT')
    except:
        pass


    #### skin that mesh!
    print('skinning mesh \o/')
    bpy.data.objects['stick_figure_mesh'].modifiers.new('stick_figure_mesh_skin', 'SKIN')
    print('resizing mesh')
    skin_radius = .025
    bpy.ops.transform.skin_resize(value=(skin_radius, skin_radius, skin_radius))

    ########
    # # parent mesh to armature
    print('parenting mesh to armarture with automatic weights...')
    bpy.ops.object.mode_set(mode='OBJECT')
    bpy.data.objects['stick_figure_mesh'].parent = this_metarig
    bpy.context.view_layer.objects.active = this_metarig
    bpy.context.active_object.select_set(True)
    bpy.data.objects['stick_figure_mesh'].select_set(True)
    bpy.ops.object.parent_set(type='ARMATURE_AUTO')


    #####################
    ## Load nSynched Videos
    try:
        print('loading videos as planes...')
        vidFolderPath = session_path / 'SyncedVideos'

        world_origin = bpy.data.objects['world_origin']

        number_of_videos = len(list(vidFolderPath.glob('*.mp4')))

        vid_location_scale = 1


        for vid_number, thisVidPath, in enumerate(vidFolderPath.glob('*.mp4')):
            print(thisVidPath)
            # use 'images as planes' add on to load in the video files as planes
            bpy.ops.import_image.to_plane(files=[{"name":thisVidPath.name}], directory=str(thisVidPath.parent), shader='EMISSION', )
            this_vid_as_plane = bpy.context.active_object
            this_vid_as_plane.name = 'vid_' + str(vid_number)
            
            
            vid_x = (vid_number - number_of_videos/2) * vid_location_scale
            
            this_vid_as_plane.location = [vid_x, vid_location_scale, vid_location_scale]
            this_vid_as_plane.rotation_euler = [np.pi/2, 0, 0]
            this_vid_as_plane.scale = [vid_location_scale*1.5]*3
            this_vid_as_plane.parent = world_origin 
            # create a light
            # bpy.ops.object.light_add(type='POINT', radius=1, align='WORLD')
    except:
        print('Failed to load videos to Blender scene - You might need to install the "images as planes" addon to this version of Blender')
except Exception as e:
    print(e)
    print('Something broke in the Blender Megascript - Maybe it failed to find a good clean frame and then that borked something downstream?')
    
# save .blend file
sessionID = session_path.stem
blend_file_save_path = session_path / (sessionID + '.blend')
print(f"Saving Blender output file to: {str(blend_file_save_path)}")
bpy.ops.wm.save_as_mainfile(filepath= str(blend_file_save_path) )


import logging
import os
import traceback
from pathlib import Path
from typing import List

import cv2
import numpy as np
import pandas as pd

from src.cameras.capture.dataclasses.frame_payload import FramePayload
from src.cameras.persistence.video_writer.save_options_dataclass import SaveOptions
from src.config.data_paths import freemocap_data_path

logger = logging.getLogger(__name__)


class VideoRecorder:
    def __init__(self,
                 camera_name: str,
                 image_width: int,
                 image_height: int,
                 fourcc: str = "MP4V"
                 ):
        self._camera_name = camera_name
        self._image_width = image_width
        self._image_height = image_height
        self._fourcc = fourcc
        self._frames: List[FramePayload] = []
        self._path_to_save_video = None
        self._video_type_str = ''
        self._timestamps_npy = np.empty(0)

    @property
    def frame_count(self):
        return len(self._frames)

    def record(self, frame_payload: FramePayload):
        self._frames.append(frame_payload)

    def save_to_disk(self, path_to_save_video: Path, video_type_str: str = 'raw'):
        if len(self._frames) == 0:
            logging.error(f"No frames to save for camera: {self._camera_name}")
            return

        self._video_type_str = video_type_str
        self._path_to_save_video = path_to_save_video
        self._path_to_save_video.mkdir(parents=True, exist_ok=True)
        self._gather_timestamps()
        self._write_video()
        self._save_timestamps()

    def _write_video(self):
        video_file_name = self._camera_name + "_synchronized_" + self._video_type_str + ".mp4"
        video_full_path = self._path_to_save_video / video_file_name

        cv2_writer = cv2.VideoWriter(
            str(video_full_path),
            cv2.VideoWriter_fourcc(*self._fourcc),
            self._median_framerate,
            (self._image_width, self._image_height))

        try:
            for frame in self._frames:
                cv2_writer.write(frame.image)

        except Exception as e:
            logger.debug("Failed during save in video writer")
            traceback.print_exc()
            raise e
        finally:
            logger.info(f"Saved video to path: {video_full_path}")
            cv2_writer.release()

    def _gather_timestamps(self):
        try:
            for frame in self._frames:
                self._timestamps_npy = np.append(self._timestamps_npy, frame.timestamp)
            self._median_framerate = (np.nanmedian(np.diff(self._timestamps_npy / 1e9))) ** -1
        except:
            logger.error("Error gathering timestamps")

    def _save_timestamps(self):
        timestamp_file_name_npy = self._camera_name + "_timestamps.npy"
        np.save(str(self._path_to_save_video / timestamp_file_name_npy), self._timestamps_npy)
        logger.info(f"Saved timestamps to path: {timestamp_file_name_npy}")

        timestamp_file_name_csv = self._camera_name + "_timestamps.csv"
        self._timestamps_npy.tofile(str(self._path_to_save_video / timestamp_file_name_csv), sep=',\n')
        logger.info(f"Saved timestamps to path: {timestamp_file_name_csv}")



import glob
import cv2
from pathlib import Path
import os


def createVideo(session):

    vidSavePath = str(session.sessionPath / "{}_outVid.mp4".format(session.sessionID))
    fps = 30
    shape = 1078, 647
    frame_array = []

    for filename in session.imOutPath.glob("*.png"):
        filename = str(filename)

        # reading each files
        img = cv2.imread(filename)
        # resized = cv2.resize(img,shape)
        # shape = img.shape[:2]
        resized = cv2.resize(img, shape)
        # inserting the frames into an image array
        frame_array.append(resized)
    out = cv2.VideoWriter(vidSavePath, cv2.VideoWriter_fourcc(*"DIVX"), fps, shape)
    for i in range(len(frame_array)):
        # writing to a image array
        out.write(frame_array[i])
    out.release()


def createBodyTrackingVideos(session):

    vidSavePath = str(session.sessionPath / "{}_outVid.mp4".format(session.sessionID))
    fps = 30
    shape = 1078, 647
    # frame_array = []

    if session.useOpenPose:

        for count,videoPath in enumerate(session.session_settings['openPose_imgPathList']):
            frame_array = []
            vidSavePath = str(
                session.openPoseDataPath / "openPoseVideo_cam{}.mp4".format(count)
            )
            videoPath = Path(videoPath)
            for filename in videoPath.glob("*.png"):

                filename = str(filename)
                # reading each files
                img = cv2.imread(filename)
                # resized = cv2.resize(img,shape)
                # shape = img.shape[:2]
                resized = cv2.resize(img, shape)
                # inserting the frames into an image array
                frame_array.append(resized)
            out = cv2.VideoWriter(
                vidSavePath, cv2.VideoWriter_fourcc(*"DIVX"), fps, shape
            )
            for i in range(len(frame_array)):
                # writing to a image array
                out.write(frame_array[i])
            out.release()




# from pathlib import Path
# import bpy

# def get_session_path(scene) -> Path:
#     abspath = bpy.path.abspath(scene.fmc_session_path) #absolute path to this omc session's data
#     pathStringParts = abspath.split('\\') #split the absolute path at the \\'s
#     scene.fmc_session_id = pathStringParts[-2] #the 2nd to last element is the 'sessionID'   
#     return Path(abspath)

# class IMPORT_SCENE_OT_session_config_load(bpy.types.Operator):
#     bl_idname = 'fmc.load_session_config'
#     bl_label = 'Load Session Config.yaml'

#     def execute(self, context):
#         #find congif.yaml file for this session 
#         import_path = get_session_path(context.scene)

#         iter = -1
#         for import_fpath in import_path.glob('*_config.yaml'):
#             iter +=1
#             assert iter < 1, 'Found Multiple *_config.yaml files, and thats not supposed to happen!'
#             bpy.types.Scene.fmc_session_path = import_fpath #path to the config gile
#             bpy.types.Scene.fmc_session_ID = import_fpath.name #name of the session 
#         return{'FINISHED'}




# -*- coding: utf-8 -*-
"""
Created on Fri Feb 26 12:42:08 2021

@author: Rontc
"""
import glob
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from ruamel.yaml import YAML

from pathlib import Path
import h5py


def parseDeepLabCut(
    session,
    config_path
):  # NOTE this is assuming both the dlc config file and dlc data h5 foles are in the same folder specified in the session.dlcDataPath, if that is not the case either dlcconfig or h5files variable need to be changed

    dlcConfig = config_path # File path for dlcConfig file
    yaml=YAML()
    
    with open(dlcConfig) as configFile:
        # Open the config file
        configList = yaml.load(configFile, Loader=yaml.FullLoader)
        for item, doc in configList.items():  # For each item in the config file
            if item == "bodyparts":  # If the item is bodyparts
                bodypartList = doc  # Set that list to a variable
            if item == "skeleton":  # If the item is skeleton
                skeletonList = doc  # Set that list to a variable

    h5files = session.dlcDataPath.glob("*.h5")  # Take all h5 files from folder

    numCams = int(session.numCams)  # Get num of cams
    numFrames = session.numFrames + 10  # Get num of frames

    numImgPoints = len(bodypartList)  # Get amount of points tracked
    dlcData_nCams_nFrames_nImgPts_XYC = np.ndarray([numCams, numFrames, numImgPoints, 3])  # Create empty array for dlc points

    nn = 0  # counter for each camera
    for data in h5files:  # Loop throuxgh each camera

        with h5py.File(data) as f:  # Open each h5 file
            fullDataGroup = f.get("df_with_missing")  # Open main h5 group
            dataTable = fullDataGroup.get("table")  # Open datatable with all DLC tracked points

            for ii in range(len(dataTable)):  # Loop through each frame
                dataFromOneFrame = dataTable[ii]  # Assign frame to varible
                idx = len(dataFromOneFrame[1]) / 3  # index for reshaping data
                sortedFrame = np.array(dataFromOneFrame[1])  # Put data in numpy array
                try:
                    dlcData_nCams_nFrames_nImgPts_XYC[nn, ii, :, :] = sortedFrame.reshape(
                        int(idx), 3
                    )  # Reshape the data into the correct form
                except:
                    print('If you get an out-of-bounds error here, you may have too many DLC files in the DLCData folder - empty it out and re-run DLC to fix')
                    #NOTE - we should fix this , i.e. by moving old DLC data to an 'oldDLCData' folder within DLCData if the user tries to run DLC when there are already files in the DLCData folder
        nn += 1

    if session.debug:
        fig = plt.figure()

        ax1 = fig.add_subplot(221)
        ax2 = fig.add_subplot(222)
        ax3 = fig.add_subplot(223)
        ax4 = fig.add_subplot(224)

        cam1im = session.firstImage_nCams_list[0]
        cam2im = session.firstImage_nCams_list[1]
        cam3im = session.firstImage_nCams_list[2]
        cam4im = session.firstImage_nCams_list[3]

        ax1.imshow(cam1im)
        ax1.plot(
            dlcData_nCams_nFrames_nImgPts_XYC[0, 0, :, 0],
            dlcData_nCams_nFrames_nImgPts_XYC[0, 0, :, 1],
        )

        ax2.imshow(cam2im)
        ax2.plot(
            dlcData_nCams_nFrames_nImgPts_XYC[1, 0, :, 0],
            dlcData_nCams_nFrames_nImgPts_XYC[1, 0, :, 1],
        )

        ax3.imshow(cam3im)
        ax3.plot(
            dlcData_nCams_nFrames_nImgPts_XYC[2, 0, :, 0],
            dlcData_nCams_nFrames_nImgPts_XYC[2, 0, :, 1],
        )

        ax4.imshow(cam4im)
        ax4.plot(
            dlcData_nCams_nFrames_nImgPts_XYC[3, 0, :, 0],
            dlcData_nCams_nFrames_nImgPts_XYC[3, 0, :, 1],
        )

        plt.show()



    return dlcData_nCams_nFrames_nImgPts_XYC



#https://matplotlib.org/stable/api/animation_api.html
from operator import is_
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from mpl_toolkits.mplot3d import axes3d
from matplotlib.widgets import Slider

fig = plt.figure()
ax3d = fig.add_subplot(211,projection="3d")
xdata, ydata , zdata = [], [], []
ln, = plt.plot([], [], [], 'r-o')

axControls = fig.add_subplot(212)

numFrames = 128
#slider
animSlider = Slider(
    ax=axControls,
    label="FrameNum",
    valmin=0,
    valmax=100,
    valinit=0,
    orientation="horizontal"
)
# Animation controls
is_manual = False # True if user has taken control of the animation
interval = 0 # ms, time between animation frames
loop_len = 0.1  # seconds per loop
scale = interval / 1000 / loop_len

def init():
    ax3d.set_xlim(0, 2*np.pi)
    ax3d.set_ylim(-1, 1)
    ax3d.set_zlim(-1, 1)
    return ln,

def update_slider(frameNum):
    global is_manual
    is_manual = True
    update_figure(frameNum)

def update_figure(frameNum):
    global is_manual
    if is_manual:
        return ln, #don't change in manual mode

    xdata.append(frameNum)
    ydata.append(np.sin(frameNum))
    zdata.append(np.sin(frameNum))
    ln.set_data_3d(xdata, ydata,zdata)
    is_manual = False # the above line called update_slider, so we need to reset this
    return ln,

def on_click(event):
    # Check where the click happened
    (xm,ym),(xM,yM) = animSlider.label.clipbox.get_points()
    if xm < event.x < xM and ym < event.y < yM:
        # Event happened within the slider, ignore since it is handled in update_slider
        return
    else:
        # user clicked somewhere else on canvas = unpause
        global is_manual
        is_manual=False

# call update function on slider value change
animSlider.on_changed(update_slider)
fig.canvas.mpl_connect('button_press_event', on_click)

ani = FuncAnimation(fig, update_figure, frames=np.linspace(0, 2*np.pi, numFrames),
                    init_func=init, blit=True)
plt.show()


import logging
import platform
import time

import numpy as np
import cv2
from pydantic import BaseModel

logger = logging.getLogger(__name__)
logger.level = logging.INFO


class NoCameraAvailableException(Exception):
    pass


class FailedFrameGrabException(Exception):
    pass


class TweakedModel(BaseModel):
    class Config:
        arbitrary_types_allowed = True


# OpenCV Implementation of interacting with a camera
class OpenCVCamera(TweakedModel):
    port_number: int = 0
    name: str = "Camera0"  # `camera{}`.format(port_number)
    exposure: int = -8
    resolution_width: int = 1280
    resolution_height: int = 720
    framerate: int = 30
    opencv_video_capture_object: cv2.VideoCapture = None
    opencv_video_writer_object: cv2.VideoWriter = None

    def initialize_video_writer(
        self,
        output_file_path: str,
    ):
        self.opencv_video_writer_object = cv2.VideoWriter(
            output_file_path,
            cv2.VideoWriter_fourcc("M", "J", "P", "G"),
            self.framerate,
            (self.resolution_width, self.resolution_height),
        )

        logger.info("Initialized video writer for {}".format(self.name))

    def connect(self) -> bool:
        if platform.system() == "Windows":
            cap_backend = cv2.CAP_DSHOW
        else:
            cap_backend = cv2.CAP_ANY

        self.opencv_video_capture_object = cv2.VideoCapture(
            self.port_number, cap_backend
        )
        success, image = self.opencv_video_capture_object.read()

        # set camera stream paramters
        self.opencv_video_capture_object.set(cv2.CAP_PROP_EXPOSURE, self.exposure)
        self.opencv_video_capture_object.set(
            cv2.CAP_PROP_FRAME_WIDTH, self.resolution_width
        )
        self.opencv_video_capture_object.set(
            cv2.CAP_PROP_FRAME_HEIGHT, self.resolution_height
        )
        self.opencv_video_capture_object.set(
            cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc("M", "J", "P", "G")
        )

        if success:
            logger.error("Camera found at port number {}".format(self.port_number))
            self.name = "Camera{}".format(self.port_number)
            return success
        else:
            # raise NoCameraAvailableException()
            logger.error(
                "Could not connect to a camera at port# {}".format(self.port_number)
            )
            return success

    def get_next_frame(self):

        timestamp_ns_pre_grab = time.time_ns()
        # Why grab not read? see -> https://stackoverflow.com/questions/57716962/difference-between-video-capture-read-and-grab
        grab_success = self.opencv_video_capture_object.grab()
        timestamp_ns_post_grab = time.time_ns()
        timestamp_ns = (timestamp_ns_pre_grab + timestamp_ns_post_grab) / 2

        if grab_success:
            success, image = self.opencv_video_capture_object.retrieve()
            # logger.info('{} successfully grabbed a frame at timestamp {}'.format(self.name, timestamp_ns/1e9))
            return success, image, timestamp_ns

        return False, None, None

    def close(self):
        self.opencv_video_capture_object.release()
        logger.info("Closed camera{}".format(self.name))


if __name__ == "__main__":
    from rich.console import Console

    console = Console()
    timestamps = []
    try:
        # Test the camera
        camera = OpenCVCamera()
        camera.connect()

        while True:
            success, image, timestamp_ns = camera.get_next_frame()
            timestamps.append(timestamp_ns / 1e9)
            if success:
                mean_fps = 1 / np.mean(np.diff(timestamps))
                console.print(
                    "{} grabbed a frame at timestamp {} : mean fps = {}".format(
                        camera.name, timestamp_ns / 1e9, mean_fps
                    )
                )
            else:
                console.print(
                    "{} failed to grab a frame at timestamp {} : mean fps = {}".format(
                        camera.name, timestamp_ns / 1e9, mean_fps
                    )
                )
            cv2.imshow(camera.name + "- Press ESC to close", image)
            exit_key = cv2.waitKey(1)
            if exit_key == 27:
                break
        cv2.destroyAllWindows()
        camera.close()
    except:
        console.print_exception()



import logging
import threading
import time
import traceback

from src.cameras.capture.dataclasses.frame_payload import FramePayload

logger = logging.getLogger(__name__)


class VideoCaptureThread(threading.Thread):
    def __init__(
            self,
            get_next_frame,
    ):
        super().__init__()
        self._is_capturing_frames = False
        self._get_next_frame = get_next_frame
        self._num_frames_processed = 0
        self._elapsed = 0
        self._frame: FramePayload = FramePayload()
        self.setDaemon(True)

    @property
    def average_fps(self):
        if self._elapsed <= 0:
            return 0
        if self._num_frames_processed <= 0:
            return 0
        return self._num_frames_processed / self._elapsed

    @property
    def latest_frame(self):
        return self._frame

    @property
    def is_capturing_frames(self):
        return self._is_capturing_frames

    def stop(self):
        self._is_capturing_frames = False

    def run(self):
        self._start_frame_loop()

    def _start_frame_loop(self, save_video=False):
        self._is_capturing_frames = True
        start = time.perf_counter_ns()
        try:
            while self._is_capturing_frames:
                self._frame = self._get_next_frame()
                self._num_frames_processed += 1
                self._elapsed = time.perf_counter_ns() - start
        except:
            logger.error("Frame loop thread exited due to error")
            traceback.print_exc()
        else:
            logger.info("Frame loop thread exited.")



import logging

import cv2
import numpy as np

from src.core_processor.fps.timestamp_manager import TimestampManager
from src.core_processor.utils.image_fps_writer import write_fps_to_image

logger = logging.getLogger(__name__)


def show_cam_window(webcam_id: str, image: np.array, fps_counter_cls: TimestampManager):
    exit_key = cv2.waitKey(1)
    if exit_key == 27:
        logger.info("ESC has been pressed.")
        return False

    write_fps_to_image(
        image,
        fps_counter_cls.median_frames_per_second(webcam_id),
    )
    cv2.imshow(webcam_id, image)
    return True



from dataclasses import dataclass

import numpy as np

from src.cameras.capture.dataclasses.frame_payload import FramePayload
from src.core_processor.camera_calibration.charuco_board_detection.dataclasses.charuco_view_data import \
    CharucoViewData


@dataclass
class CharucoFramePayload:
    raw_frame_payload: FramePayload = FramePayload()
    annotated_image: np.ndarray = None
    charuco_view_data: CharucoViewData = CharucoViewData()



from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from freemocap.fmc_runme import RunMe
from freemocap import fmc_demo, fmc_runme

# from freemocap import fmc_runme, fmc_demo #JSM NOTE - there's gotta be a better way to do this lol

# def RunMe(sessionID=None,
#             stage=1,
#             useOpenPose=True, 
#             openPoseDummyRun=False,
#             useMediaPipe=False,
#             useDLC=False,
#             dlcConfigPath=None,
#             debug=False,
#             setDataPath=False,
#             userDataPath = None, 
#             recordVid=False,
#             reconstructionConfidenceThreshold = .8,
#             charucoSquareSize = 36,
#             ):
#     """ 
#     Passes user-set values from the freemocap_runme_script to the fmc_runme.Run function (and allows the user to run the pipeline with just an fmc.RunMe command)
#     """  
#     fmc_runme.Run(sessionID,
#     stage,
#     useOpenPose,
#     openPoseDummyRun, 
#     useMediaPipe,
#     useDLC,
#     dlcConfigPath,
#     debug,
#     setDataPath,
#     userDataPath,
#     recordVid)

def RunDemo():
    """ 
    Will (eventually) download sample data from Figshare, and run the animation for it
    """  
    sample_data_location, sample_data_name = fmc_demo.DemoSetup() #run a bunch of GUIs, get the location of the directory where the data folder with the sample data is 
    #fmc_runme.RunMe(sessionID = sample_data_name,stage = 5,useOpenPose = False, useMediaPipe = True,useDLC = True, userDataPath = sample_data_location)


from dataclasses import dataclass

import numpy as np


@dataclass
class CameraCalibrationData:
    image_width: int
    image_height: int
    reprojection_error: float = 1e9
    camera_matrix: np.ndarray = None
    number_of_lens_distortion_coefficients: int = 4
    lens_distortion_coefficients: np.ndarray = None
    rotation_vectors_of_the_board: np.ndarray = None
    translation_vectors_of_the_board: np.ndarray = None
    lens_distortion_std_dev: float = None
    camera_location_std_dev: float = None

    def __post_init__(self):
        if self.camera_matrix is None:
            self.camera_matrix = np.array([[float(self.image_width), 0., self.image_width / 2.],
                                           [0., float(self.image_width), self.image_height / 2.],
                                           [0., 0., 1.]])

        if self.lens_distortion_coefficients is None:
            self.lens_distortion_coefficients = np.zeros((self.number_of_lens_distortion_coefficients, 1))


